
import base64
import datetime
import functools
import io
import os
import typing
import urllib.parse

import flask
import flask_login
import markupsafe
import qrcode
import qrcode.image.pil
from bs4 import BeautifulSoup
from flask_babel import _, refresh
from weasyprint import HTML

from .. import logic
from ..logic import object_log, object_permissions
from ..logic.actions import Action
from ..logic.objects import get_object
from ..models import ObjectLogEntryType, Permissions
from ..logic.users import get_user

from .markdown_images import IMAGE_FORMATS
from .utils import custom_format_datetime, get_user_if_exists, get_location_name
from ..logic.utils import get_translated_text
from ..logic.caching import cache_per_request

SECTIONS = frozenset({
    'activity_log',
    'locations',
    'publications',
    'files',
    'comments'
})


def create_html_for_pdfexport(
        user_id: int,
        object_ids: typing.Sequence[int],
        sections: typing.Union[typing.Set[str], typing.FrozenSet[str]] = SECTIONS,
        lang_code: str = 'en',
        *,
        url_mapper: typing.Callable[[str, typing.Dict[typing.Tuple[int, int], logic.files.File]], str],
        include_logo: bool = True,
) -> str:
    exported_files: typing.Dict[typing.Tuple[int, int], logic.files.File] = {}

    flask.g.override_locale = lang_code
    refresh()

    base_url = flask.url_for('.index', _external=True)

    objects = []
    for object_id in object_ids:
        object = get_object(object_id)

        activity_log_entries = []
        if 'activity_log' in sections:
            object_log_entries = object_log.get_object_log_entries(object_id=object.id, user_id=user_id)
            for object_log_entry in reversed(object_log_entries):
                user_url = markupsafe.escape(flask.url_for('.user_profile', user_id=object_log_entry.user_id, _external=True))
                user_name = markupsafe.escape(get_user(object_log_entry.user_id).get_name())

                entry_datetime = markupsafe.escape(custom_format_datetime(object_log_entry.utc_datetime))
                text = f'{entry_datetime} — '
                if object_log_entry.type == ObjectLogEntryType.CREATE_BATCH:
                    text += _('<a href="%(user_url)s">%(user_name)s</a> created this object as part of a batch.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.CREATE_OBJECT:
                    text += _('<a href="%(user_url)s">%(user_name)s</a> created this object.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.EDIT_OBJECT:
                    text += _('<a href="%(user_url)s">%(user_name)s</a> edited this object.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.POST_COMMENT:
                    text += _('<a href="%(user_url)s">%(user_name)s</a> posted a comment.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.RESTORE_OBJECT_VERSION:
                    text += _('<a href="%(user_url)s">%(user_name)s</a> restored a previous version of this object.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.UPLOAD_FILE:
                    text += _('<a href="%(user_url)s">%(user_name)s</a> posted a file.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.USE_OBJECT_IN_MEASUREMENT:
                    try:
                        measurement_id = int(object_log_entry.data['measurement_id'])
                        escaped_object_url = markupsafe.escape(flask.url_for('.object', object_id=measurement_id, _external=True))
                        permissions = logic.object_permissions.get_user_object_permissions(measurement_id, user_id)
                        if Permissions.READ in permissions:
                            measurement_name = markupsafe.escape(get_translated_text(get_object(measurement_id).name))
                            text += _('<a href="%(user_url)s">%(user_name)s</a> used this object in <a href="%(object_url)s">measurement %(measurement_name)s (#%(measurement_id)s)</a>.', user_url=user_url, user_name=user_name, object_url=escaped_object_url, measurement_id=measurement_id, measurement_name=measurement_name)
                        else:
                            text += _('<a href="%(user_url)s">%(user_name)s</a> used this object in <a href="%(object_url)s">measurement #%(measurement_id)s</a>.', user_url=user_url, user_name=user_name, object_url=escaped_object_url, measurement_id=measurement_id)
                    except Exception:
                        text += _('<a href="%(user_url)s">%(user_name)s</a> used this object in a measurement.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.USE_OBJECT_IN_SAMPLE_CREATION:
                    try:
                        sample_id = int(object_log_entry.data['sample_id'])
                        escaped_object_url = markupsafe.escape(flask.url_for('.object', object_id=sample_id, _external=True))
                        permissions = logic.object_permissions.get_user_object_permissions(sample_id, user_id)
                        if Permissions.READ in permissions:
                            sample_name = markupsafe.escape(get_translated_text(get_object(sample_id).name))
                            text += _('<a href="%(user_url)s">%(user_name)s</a> used this object to create <a href="%(object_url)s">sample %(sample_name)s (#%(sample_id)s)</a>.', user_url=user_url, user_name=user_name, object_url=escaped_object_url, sample_id=sample_id, sample_name=sample_name)
                        else:
                            text += _('<a href="%(user_url)s">%(user_name)s</a> used this object to create <a href="%(object_url)s">sample #%(sample_id)s</a>.', user_url=user_url, user_name=user_name, object_url=escaped_object_url, sample_id=sample_id)
                    except Exception:
                        text += _('<a href="%(user_url)s">%(user_name)s</a> used this object to create a sample.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.ASSIGN_LOCATION:
                    object_location_assignment_id = object_log_entry.data['object_location_assignment_id']
                    object_location_assignment = logic.locations.get_object_location_assignment(object_location_assignment_id)
                    if object_location_assignment.location_id is not None:
                        location_url = markupsafe.escape(
                            flask.url_for(
                                '.location',
                                location_id=object_location_assignment.location_id,
                                _external=True
                            )
                        )
                        location_name = get_location_name(
                            location_or_location_id=object_location_assignment.location_id,
                            include_id=True,
                            language_code=lang_code
                        )
                    else:
                        location_url = None
                        location_name = None
                    if object_location_assignment.responsible_user_id is not None:
                        other_user_url = markupsafe.escape(
                            flask.url_for(
                                '.user_profile',
                                user_id=object_location_assignment.responsible_user_id,
                                _external=True
                            )
                        )
                    else:
                        other_user_url = None
                    if object_location_assignment.confirmed:
                        responsibility_status = _(' (confirmed)')
                    elif object_location_assignment.declined:
                        responsibility_status = _(' (declined)')
                    else:
                        responsibility_status = _(' (unconfirmed)')
                    if object_location_assignment.location_id is not None and object_location_assignment.responsible_user_id is not None:
                        text += _(
                            '<a href="%(user_url)s">%(user_name)s</a> assigned this object to <a href="%(location_url)s">%(location_name)s</a> and <a href="%(other_user_url)s">user #%(responsible_user_id)s</a>%(responsibility_status)s.',
                            user_url=user_url,
                            user_name=user_name,
                            location_url=location_url,
                            location_name=location_name,
                            other_user_url=other_user_url,
                            responsible_user_id=object_location_assignment.responsible_user_id,
                            responsibility_status=responsibility_status
                        )
                    elif object_location_assignment.location_id is not None:
                        text += _(
                            '<a href="%(user_url)s">%(user_name)s</a> assigned this object to <a href="%(location_url)s">%(location_name)s</a>.',
                            user_url=user_url,
                            user_name=user_name,
                            location_url=location_url,
                            location_name=location_name
                        )
                    elif object_location_assignment.responsible_user_id is not None:
                        text += _(
                            '<a href="%(user_url)s">%(user_name)s</a> assigned this object to <a href="%(other_user_url)s">user #%(responsible_user_id)s</a>%(responsibility_status)s.',
                            user_url=user_url,
                            user_name=user_name,
                            other_user_url=other_user_url,
                            responsible_user_id=object_location_assignment.responsible_user_id,
                            responsibility_status=responsibility_status
                        )
                elif object_log_entry.type == ObjectLogEntryType.LINK_PUBLICATION:
                    doi = markupsafe.escape(object_log_entry.data['doi'])
                    text += _('<a href="%(user_url)s">%(user_name)s</a> linked publication <a href="https://dx.doi.org/%(doi)s">%(doi)s</a> to this object.', user_url=user_url, user_name=user_name, doi=doi)
                elif object_log_entry.type == ObjectLogEntryType.REFERENCE_OBJECT_IN_METADATA:
                    if object_log_entry.data['object_id'] is not None:
                        try:
                            other_object_id = int(object_log_entry.data['object_id'])
                            escaped_object_url = markupsafe.escape(flask.url_for('.object', object_id=other_object_id, _external=True))
                            permissions = logic.object_permissions.get_user_object_permissions(object_log_entry.data['object_id'], user_id)
                            if Permissions.READ in permissions:
                                object_name = markupsafe.escape(get_translated_text(get_object(object_log_entry.data['object_id']).name))
                                text += _('<a href="%(user_url)s">%(user_name)s</a> referenced this object in the metadata of <a href="%(object_url)s">object %(object_name)s (#%(other_object_id)s)</a>.', user_url=user_url, user_name=user_name, object_url=escaped_object_url, object_name=object_name, other_object_id=other_object_id)
                            else:
                                text += _('<a href="%(user_url)s">%(user_name)s</a> referenced this object in the metadata of <a href="%(object_url)s">object #%(other_object_id)s</a>.', user_url=user_url, user_name=user_name, object_url=escaped_object_url, other_object_id=other_object_id)
                        except Exception:
                            text += _('<a href="%(user_url)s">%(user_name)s</a> referenced this object in the metadata of an unknown object.', user_url=user_url, user_name=user_name)
                    else:
                        text += _('<a href="%(user_url)s">%(user_name)s</a> referenced this object in the metadata of another object.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.EXPORT_TO_DATAVERSE:
                    dataverse_url = markupsafe.escape(object_log_entry.data['dataverse_url'])
                    text += _('<a href="%(user_url)s">%(user_name)s</a> exported this object to dataverse as <a href="%(dataverse_url)s">%(dataverse_url)s</a>.', user_url=user_url, user_name=user_name, dataverse_url=dataverse_url)
                elif object_log_entry.type == ObjectLogEntryType.LINK_PROJECT:
                    try:
                        project = logic.projects.get_project(object_log_entry.data['project_id'])
                        project_url = flask.url_for('.project', project_id=project.id)
                        text += _('<a href="%(user_url)s">%(user_name)s</a> linked this object to <a href="%(project_url)s">project group %(project_name)s (#%(project_id)s)</a>.', user_url=user_url, user_name=user_name, project_url=project_url, project_name=get_translated_text(project.name), project_id=project.id)
                    except logic.errors.ProjectDoesNotExistError:
                        text += _('<a href="%(user_url)s">%(user_name)s</a> linked this object to a project group.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.IMPORT_FROM_ELN_FILE:
                    text += _('<a href="%(user_url)s">%(user_name)s</a> imported this object from an .eln file.', user_url=user_url, user_name=user_name)
                elif object_log_entry.type == ObjectLogEntryType.UNLINK_PROJECT:
                    if object_log_entry.data.get('project_deleted'):
                        text += _('<a href="%(user_url)s">%(user_name)s</a> deleted the project group this object was linked to.', user_url=user_url, user_name=user_name)
                    else:
                        try:
                            project = logic.projects.get_project(object_log_entry.data['project_id'])
                            project_url = flask.url_for('.project', project_id=project.id)
                            text += _('<a href="%(user_url)s">%(user_name)s</a> removed the link of this object to <a href="%(project_url)s">project group %(project_name)s (#%(project_id)s)</a>.', user_url=user_url, user_name=user_name, project_url=project_url, project_name=get_translated_text(project.name), project_id=project.id)
                        except logic.errors.ProjectDoesNotExistError:
                            text += _('<a href="%(user_url)s">%(user_name)s</a> removed the link of this object to a project group.', user_url=user_url, user_name=user_name)
                else:
                    text += _('<a href="%(user_url)s">%(user_name)s</a> performed an unknown action.', user_url=user_url, user_name=user_name)
                activity_log_entries.append(text)

        locations_entries = []
        if 'locations' in sections:
            location_assignments = logic.locations.get_object_location_assignments(object.id)
            for location_assignment in location_assignments:
                locations_entries.append({
                    'utc_datetime': location_assignment.utc_datetime.strftime('%Y-%m-%d %H:%M') if location_assignment.utc_datetime is not None else '',
                    'assigning_user_id': location_assignment.user_id,
                    'assigning_user_name': logic.users.get_user(location_assignment.user_id).get_name() if location_assignment.user_id is not None else '',
                    'location_id': location_assignment.location_id,
                    'location_name': get_location_name(location_assignment.location_id, include_id=True, language_code=lang_code) if location_assignment.location_id else None,
                    'responsible_user_id': location_assignment.responsible_user_id,
                    'responsible_user_name': logic.users.get_user(location_assignment.responsible_user_id).get_name() if location_assignment.responsible_user_id else None,
                    'description': location_assignment.description
                })

        if 'publications' in sections:
            publications = logic.publications.get_publications_for_object(object.id)
        else:
            publications = []

        if 'files' in sections:
            files = logic.files.get_files_for_object(object.id)
            for file in files:
                exported_files[(object.id, file.id)] = file
        else:
            files = []

        if 'comments' in sections:
            comments = logic.comments.get_comments_for_object(object.id)
        else:
            comments = []

        object_url = flask.url_for('.object', object_id=object_id, _external=True)
        image = qrcode.make(object_url, image_factory=qrcode.image.pil.PilImage)
        qrcode_width, qrcode_height = image.size
        # remove margin
        margin = 40
        image = image.crop((margin, margin, qrcode_width - margin, qrcode_height - margin))
        qrcode_file = io.BytesIO()
        image.save(qrcode_file, format='PNG')
        qrcode_file.seek(0)
        qrcode_url = 'data:image/png;base64,' + base64.b64encode(qrcode_file.read()).decode('utf-8')

        if object.action_id is not None:
            action = logic.actions.get_action(object.action_id)
        else:
            action = None
        objects.append((object, action, activity_log_entries, locations_entries, publications, comments, files, qrcode_url))

    def get_object_type_name(action: Action) -> str:
        if action is None or action.type is None:
            return _('Object')
        else:
            return get_translated_text(action.type.object_name, default=_('Object'))

    files_by_object_id = {}
    for object_tuple in objects:
        files_by_object_id[object_tuple[0].object_id] = logic.files.get_files_for_object(object_tuple[0].object_id)
    html = flask.render_template(
        'pdfexport/export.html',
        get_object_type_name=get_object_type_name,
        export_date=datetime.datetime.now(datetime.timezone.utc),
        get_object_if_current_user_has_read_permissions=cache_per_request()(functools.partial(object_permissions.get_object_if_user_has_permissions, user_id, Permissions.READ)),
        objects=objects,
        get_user=get_user_if_exists,
        metadata_language=lang_code,
        files_by_object_id=files_by_object_id,
        eln_import_urls={object[0].object_id: logic.eln_import.get_eln_import_object_url(object[0].object_id) for object in objects},
        IMAGE_FORMATS=IMAGE_FORMATS,
        include_logo=include_logo,
    )

    soup = BeautifulSoup(html, 'html.parser')
    for tag_name, url_attribute in [('a', 'href'), ('img', 'src')]:
        for element in soup.findAll(tag_name):
            url = element.get(url_attribute)
            if url and not urllib.parse.urlparse(url).netloc:
                element[url_attribute] = url_mapper(urllib.parse.urljoin(base_url, url), exported_files)
    html = str(soup)

    # use regular user language again
    delattr(flask.g, 'override_locale')
    refresh()

    return html


def create_pdfexport(
        object_ids: typing.Sequence[int],
        sections: typing.Union[typing.Set[str], typing.FrozenSet[str]] = SECTIONS,
        lang_code: str = 'en'
) -> bytes:
    base_url = flask.url_for('.index', _external=True)

    def url_mapper(url: str, exported_files: typing.Dict[typing.Tuple[int, int], logic.files.File]) -> str:
        # replace URLs of markdown images with Data URLs
        if url.startswith(base_url + 'markdown_images/'):
            file_name = url[len(base_url + 'markdown_images/'):]
            image_data = logic.markdown_images.get_markdown_image(file_name, flask_login.current_user.id)
            if image_data is None:
                url = ''
            else:
                file_extension = os.path.splitext(file_name)[1].lower()
                if file_extension in IMAGE_FORMATS:
                    url = 'data:' + IMAGE_FORMATS[file_extension] + ';base64,' + base64.b64encode(image_data).decode('utf-8')
                else:
                    url = ''
        if url.startswith(base_url + 'objects/') and '/files/' in url[len(base_url + 'objects/'):]:
            object_id_file_id = url[len(base_url + 'objects/'):]
            url = ''
            try:
                object_id_str, file_id_str = object_id_file_id.split('/files/')
                object_id = int(object_id_str)
                file_id = int(file_id_str)
                file = exported_files[(object_id, file_id)]
                if file.storage == 'database' and not file.is_hidden:
                    for file_extension, mime_type in IMAGE_FORMATS.items():
                        if file.original_file_name.lower().endswith(file_extension):
                            image_data = file.open(read_only=True).read()
                            url = 'data:' + mime_type + ';base64,' + base64.b64encode(image_data).decode('utf-8')
                            break
            except Exception:
                pass
        # only allow Data URLs and URLs via http or https
        if not (url.startswith('data:') or urllib.parse.urlparse(url).scheme in ('http', 'https')):
            url = ''
        return url

    html = create_html_for_pdfexport(
        user_id=flask_login.current_user.id,
        object_ids=object_ids,
        sections=sections,
        lang_code=lang_code,
        url_mapper=url_mapper
    )
    return typing.cast(bytes, HTML(
        string=html,
        base_url=base_url
    ).write_pdf())



import typing

import flask

from . import core
from ... import logic


def post_poke_components_task() -> None:
    if flask.current_app.config["ENABLE_BACKGROUND_TASKS"]:
        core.post_background_task(
            type='poke_components',
            data={},
            auto_delete=True
        )
    else:
        handle_poke_components_task({}, None)


def handle_poke_components_task(
    data: typing.Dict[str, typing.Any],
    task_id: typing.Optional[int]
) -> typing.Tuple[bool, typing.Optional[dict[str, typing.Any]]]:
    for component in logic.components.get_components():
        if component.export_token_available:
            try:
                logic.federation.update.update_poke_component(component)
            except Exception:
                pass
    return True, {}



import importlib.util
import os.path
from setuptools import setup

setup_directory = os.path.abspath(os.path.dirname(__file__))

version_file_path = os.path.join(setup_directory, 'sampledb', 'version.py')
spec = importlib.util.spec_from_file_location("version", version_file_path)
version = importlib.util.module_from_spec(spec)
spec.loader.exec_module(version)

setup(
    version=version.__version__,
)



# coding: utf-8
"""
Add user_id column to fed_object_log_entries table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('fed_object_log_entries', 'user_id'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE fed_object_log_entries
        ADD user_id INTEGER REFERENCES users(id)
    """))
    return True



# coding: utf-8
"""
Script for listing all actions in SampleDB.

Usage: sampledb list_actions
"""
import sys
import typing

from .. import create_app
from ..logic.actions import get_actions


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 0:
        print(__doc__)
        sys.exit(1)
    app = create_app()
    with app.app_context():
        actions = get_actions()
        for action in actions:
            print(f" - #{action.id}: {action.name.get('en', 'Unnamed Action')}")



# coding: utf-8
"""
Add create_log_entry_default column to instruments table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('instruments', 'create_log_entry_default'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE instruments
        ADD create_log_entry_default BOOLEAN NOT NULL DEFAULT FALSE
    """))
    return True



# coding: utf-8
"""
RESTful API for SampleDB
"""

from flask import Blueprint
from flask_cors import CORS
from .objects import Object, Objects, ObjectVersion, ObjectVersions, RelatedObjects
from .actions import Action, Actions
from .action_types import ActionType, ActionTypes
from .authentication import AccessTokens
from .comments import ObjectComment, ObjectComments
from .files import ObjectFile, ObjectFiles
from .instruments import Instrument, Instruments
from .instrument_log import InstrumentLogEntry, InstrumentLogEntries, InstrumentLogEntryVersions, InstrumentLogEntryVersion, InstrumentLogEntryFileAttachment, InstrumentLogEntryFileAttachments, InstrumentLogEntryObjectAttachment, InstrumentLogEntryObjectAttachments, InstrumentLogCategory, InstrumentLogCategories
from .locations import Location, Locations, ObjectLocationAssignment, ObjectLocationAssignments, LocationType, LocationTypes
from .object_log import ObjectLogEntries
from .object_permissions import UsersObjectPermissions, UserObjectPermissions, GroupsObjectPermissions, GroupObjectPermissions, ProjectsObjectPermissions, ProjectObjectPermissions, PublicObjectPermissions, AuthenticatedUserObjectPermissions, AnonymousUserObjectPermissions
from .users import CurrentUser, User, Users
from .groups import Group, Groups
from .projects import Project, Projects

api = Blueprint('api', __name__)
CORS(api)
api.add_url_rule('/api/v1/objects/', endpoint='objects', view_func=Objects.as_view('objects'))
api.add_url_rule('/api/v1/objects/<int:object_id>', endpoint='object', view_func=Object.as_view('object'))
api.add_url_rule('/api/v1/objects/<int:object_id>/versions/', endpoint='object_versions', view_func=ObjectVersions.as_view('object_versions'))
api.add_url_rule('/api/v1/objects/<int:object_id>/versions/<int:version_id>', endpoint='object_version', view_func=ObjectVersion.as_view('object_version'))
api.add_url_rule('/api/v1/objects/<int:object_id>/related_objects', endpoint='related_objects', view_func=RelatedObjects.as_view('related_objects'))
api.add_url_rule('/api/v1/actions/', endpoint='actions', view_func=Actions.as_view('actions'))
api.add_url_rule('/api/v1/actions/<int:action_id>', endpoint='action', view_func=Action.as_view('action'))
api.add_url_rule('/api/v1/action_types/', endpoint='action_types', view_func=ActionTypes.as_view('action_types'))
api.add_url_rule('/api/v1/action_types/<int(signed=True):type_id>', endpoint='action_type', view_func=ActionType.as_view('action_type'))
api.add_url_rule('/api/v1/access_tokens/', endpoint='access_tokens', view_func=AccessTokens.as_view('access_tokens'))
api.add_url_rule('/api/v1/objects/<int:object_id>/comments/', endpoint='object_comments', view_func=ObjectComments.as_view('object_comments'))
api.add_url_rule('/api/v1/objects/<int:object_id>/comments/<int:comment_id>', endpoint='object_comment', view_func=ObjectComment.as_view('object_comment'))
api.add_url_rule('/api/v1/instruments/', endpoint='instruments', view_func=Instruments.as_view('instruments'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>', endpoint='instrument', view_func=Instrument.as_view('instrument'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_entries/', endpoint='instrument_log_entries', view_func=InstrumentLogEntries.as_view('instrument_log_entries'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>', endpoint='instrument_log_entry', view_func=InstrumentLogEntry.as_view('instrument_log_entry'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/versions/', endpoint='instrument_log_entry_versions', view_func=InstrumentLogEntryVersions.as_view('instrument_log_entry_versions'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/versions/<int:log_entry_version_id>', endpoint='instrument_log_entry_version', view_func=InstrumentLogEntryVersion.as_view('instrument_log_entry_version'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/file_attachments/', endpoint='instrument_log_entry_file_attachments', view_func=InstrumentLogEntryFileAttachments.as_view('instrument_log_entry_file_attachments'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/file_attachments/<int:file_attachment_id>', endpoint='instrument_log_entry_file_attachment', view_func=InstrumentLogEntryFileAttachment.as_view('instrument_log_entry_file_attachment'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/object_attachments/', endpoint='instrument_log_entry_object_attachments', view_func=InstrumentLogEntryObjectAttachments.as_view('instrument_log_entry_object_attachments'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_entries/<int:log_entry_id>/object_attachments/<int:object_attachment_id>', endpoint='instrument_log_entry_object_attachment', view_func=InstrumentLogEntryObjectAttachment.as_view('instrument_log_entry_object_attachment'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_categories/', endpoint='instrument_log_categories', view_func=InstrumentLogCategories.as_view('instrument_log_categories'))
api.add_url_rule('/api/v1/instruments/<int:instrument_id>/log_categories/<int:category_id>', endpoint='instrument_log_category', view_func=InstrumentLogCategory.as_view('instrument_log_category'))
api.add_url_rule('/api/v1/locations/', endpoint='locations', view_func=Locations.as_view('locations'))
api.add_url_rule('/api/v1/locations/<int:location_id>', endpoint='location', view_func=Location.as_view('location'))
api.add_url_rule('/api/v1/location_types/', endpoint='location_types', view_func=LocationTypes.as_view('location_types'))
api.add_url_rule('/api/v1/location_types/<int(signed=True):location_type_id>', endpoint='location_type', view_func=LocationType.as_view('location_type'))
api.add_url_rule('/api/v1/objects/<int:object_id>/files/', endpoint='object_files', view_func=ObjectFiles.as_view('object_files'))
api.add_url_rule('/api/v1/objects/<int:object_id>/files/<int:file_id>', endpoint='object_file', view_func=ObjectFile.as_view('object_file'))
api.add_url_rule('/api/v1/objects/<int:object_id>/locations/', endpoint='object_location_assignments', view_func=ObjectLocationAssignments.as_view('object_location_assignments'))
api.add_url_rule('/api/v1/objects/<int:object_id>/locations/<int:object_location_assignment_index>', endpoint='object_location_assignment', view_func=ObjectLocationAssignment.as_view('object_location_assignment'))
api.add_url_rule('/api/v1/objects/<int:object_id>/permissions/users/', endpoint='users_object_permissions', view_func=UsersObjectPermissions.as_view('users_object_permissions'))
api.add_url_rule('/api/v1/objects/<int:object_id>/permissions/users/<int:user_id>', endpoint='user_object_permissions', view_func=UserObjectPermissions.as_view('user_object_permissions'))
api.add_url_rule('/api/v1/objects/<int:object_id>/permissions/groups/', endpoint='groups_object_permissions', view_func=GroupsObjectPermissions.as_view('groups_object_permissions'))
api.add_url_rule('/api/v1/objects/<int:object_id>/permissions/groups/<int:group_id>', endpoint='group_object_permissions', view_func=GroupObjectPermissions.as_view('group_object_permissions'))
api.add_url_rule('/api/v1/objects/<int:object_id>/permissions/projects/', endpoint='projects_object_permissions', view_func=ProjectsObjectPermissions.as_view('projects_object_permissions'))
api.add_url_rule('/api/v1/objects/<int:object_id>/permissions/projects/<int:project_id>', endpoint='project_object_permissions', view_func=ProjectObjectPermissions.as_view('project_object_permissions'))
api.add_url_rule('/api/v1/objects/<int:object_id>/permissions/public', endpoint='public_object_permissions', view_func=PublicObjectPermissions.as_view('public_object_permissions'))
api.add_url_rule('/api/v1/objects/<int:object_id>/permissions/authenticated_users', endpoint='all_user_object_permissions', view_func=AuthenticatedUserObjectPermissions.as_view('all_user_object_permissions'))
api.add_url_rule('/api/v1/objects/<int:object_id>/permissions/anonymous_users', endpoint='anonymous_user_object_permissions', view_func=AnonymousUserObjectPermissions.as_view('anonymous_user_object_permissions'))
api.add_url_rule('/api/v1/users/', endpoint='users', view_func=Users.as_view('users'))
api.add_url_rule('/api/v1/users/<int:user_id>', endpoint='user', view_func=User.as_view('user'))
api.add_url_rule('/api/v1/users/me', endpoint='current_user', view_func=CurrentUser.as_view('current_user'))
api.add_url_rule('/api/v1/groups/', endpoint='groups', view_func=Groups.as_view('groups'))
api.add_url_rule('/api/v1/groups/<int:group_id>', endpoint='group', view_func=Group.as_view('group'))
api.add_url_rule('/api/v1/projects/', endpoint='projects', view_func=Projects.as_view('projects'))
api.add_url_rule('/api/v1/projects/<int:project_id>', endpoint='project', view_func=Project.as_view('project'))
api.add_url_rule('/api/v1/object_log_entries/', endpoint='object_log_entries', view_func=ObjectLogEntries.as_view('object_log_entry'))



# coding: utf-8
"""
Add content_is_markdown column to instrument_log_entry_versions table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('instrument_log_entry_versions', 'content_is_markdown'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE instrument_log_entry_versions
        ADD content_is_markdown BOOLEAN DEFAULT FALSE NOT NULL
    """))
    return True



# coding: utf-8
"""

"""
import copy
import secrets
import typing
from copy import deepcopy
import datetime

import flask
import flask_login
import itsdangerous
from flask_babel import _

from ... import logic
from ... import models
from ...logic.actions import Action
from ...logic.action_permissions import get_sorted_actions_for_user
from ...logic.object_permissions import get_user_object_permissions
from ...logic.users import get_users
from ...logic.schemas import generate_placeholder
from ...logic.objects import create_object, create_object_batch, update_object
from ...logic.languages import get_language, get_languages, Language
from ...logic.components import get_component
from .forms import ObjectForm
from ..utils import default_format_datetime, custom_format_number, format_time
from .object_form_parser import parse_form_data
from ...logic.utils import get_translated_text
from .permissions import get_object_if_current_user_has_read_permissions
from ...models import Permissions, Object
from ...utils import FlaskResponseT


def show_object_form(
        object: typing.Optional[Object],
        action: Action,
        previous_object: typing.Optional[Object] = None,
        should_upgrade_schema: bool = False,
        placeholder_data: typing.Optional[typing.Dict[typing.Sequence[typing.Union[str, int]], typing.Any]] = None,
        possible_object_id_properties: typing.Optional[typing.Dict[str, typing.Any]] = None,
        passed_object_ids: typing.Optional[typing.List[int]] = None,
        show_selecting_modal: bool = False
) -> FlaskResponseT:

    if object is None:
        if action.type is None or action.type.disable_create_objects or action.disable_create_objects or (action.admin_only and not flask_login.current_user.is_admin):
            flask.flash(_('Creating objects with this action has been disabled.'), 'error')
            return flask.redirect(flask.url_for('.action', action_id=action.id))

    template_arguments: typing.Dict[str, typing.Any] = {
        'ENGLISH': get_language(Language.ENGLISH),
        'get_component': get_component,
        'get_object_if_current_user_has_read_permissions': get_object_if_current_user_has_read_permissions,
        'template_mode': 'form',
    }

    errors: typing.Dict[str, str] = {}
    form = ObjectForm()
    if flask.request.method != 'GET' and form.validate_on_submit():
        raw_form_data = {key: flask.request.form.getlist(key) for key in flask.request.form}
        form_data = {k: v[0] for k, v in raw_form_data.items()}
    else:
        raw_form_data = {}
        form_data = {}
    template_arguments.update({
        'errors': errors,
        'form': form,
    })

    template_arguments.update({
        'previous_schema': None,
        'diff': None
    })
    if should_upgrade_schema and previous_object is None:
        # edit object with schema upgrade
        mode = 'upgrade'
        if action.schema is None or object is None or object.schema is None or object.data is None:
            return flask.abort(400)
        schema = action.schema
        data, upgrade_warnings = logic.schemas.convert_to_schema(object.data, object.schema, action.schema)
        for upgrade_warning in upgrade_warnings:
            flask.flash(upgrade_warning, 'warning')
        template_arguments.update({
            'previous_schema': object.schema,
            'diff': logic.schemas.calculate_diff(object.data, data)
        })
    elif object is not None:
        # edit object
        mode = 'edit'
        if object.data is None or object.schema is None:
            return flask.abort(400)
        schema = object.schema
        data = copy.deepcopy(object.data)
    elif previous_object is not None:
        # create object via 'Use as Template'
        mode = 'create'
        if previous_object.data is None or previous_object.schema is None:
            return flask.abort(400)
        if should_upgrade_schema:
            if action.schema is None:
                return flask.abort(400)
            schema = action.schema
            data, upgrade_warnings = logic.schemas.convert_to_schema(previous_object.data, previous_object.schema, action.schema)
            for upgrade_warning in upgrade_warnings:
                flask.flash(upgrade_warning, 'warning')
            template_arguments.update({
                'previous_schema': previous_object.schema,
                'diff': logic.schemas.calculate_diff(previous_object.data, data)
            })
        else:
            schema = previous_object.schema
            data = logic.schemas.copy_data(previous_object.data, previous_object.schema)
        template_arguments.update({
            'previous_object_id': previous_object.id,
            'has_grant_for_previous_object': Permissions.GRANT in get_user_object_permissions(user_id=flask_login.current_user.id, object_id=previous_object.id),
        })
    else:
        # create object from scratch
        mode = 'create'
        if action.schema is None:
            return flask.abort(400)
        schema = action.schema
        data = generate_placeholder(action.schema)
        if not isinstance(data, dict):
            return flask.abort(400)
        if placeholder_data:
            _apply_placeholder_data(placeholder_data, data)
        template_arguments.update({
            'previous_object_id': None,
            'has_grant_for_previous_object': False,
        })

    if passed_object_ids:
        has_grant_for_first_passed_object = Permissions.GRANT in get_user_object_permissions(user_id=flask_login.current_user.id, object_id=passed_object_ids[0])
    else:
        has_grant_for_first_passed_object = False
    template_arguments.update({
        'has_grant_for_first_passed_object': has_grant_for_first_passed_object,
    })

    if not isinstance(data, dict):
        return flask.abort(400)
    if not isinstance(schema, dict):
        return flask.abort(400)

    template_arguments.update({
        'data': data,
        'schema': schema,
    })

    if object is None:
        # allow creating an instrument log entry for a newly created object
        if not flask.current_app.config['DISABLE_INSTRUMENTS'] and action is not None and action.instrument is not None and flask_login.current_user in action.instrument.responsible_users:
            may_create_log_entry = True
            create_log_entry_default = action.instrument.create_log_entry_default
            instrument_log_categories = logic.instrument_log_entries.get_instrument_log_categories(action.instrument.id)
        else:
            may_create_log_entry = False
            create_log_entry_default = False
            instrument_log_categories = []
        template_arguments.update({
            'may_create_log_entry': may_create_log_entry,
            'create_log_entry_default': create_log_entry_default,
            'instrument_log_categories': instrument_log_categories,
        })

    template_arguments.update({
        'form_data': form_data,
    })

    template_arguments.update(get_object_form_template_kwargs(object.id if object is not None else None))

    actual_file_names_by_id = {
        file_id: file_names[0]
        for file_id, file_names in template_arguments['file_names_by_id'].items()
    }
    context_id = template_arguments['context_id']

    if "action_submit" in form_data:
        batch_names = _handle_batch_names(schema, form_data, raw_form_data, errors)
        object_data, parsing_errors = parse_form_data(raw_form_data, schema, file_names_by_id=actual_file_names_by_id, previous_data=data)
        errors.update(parsing_errors)
        if batch_names is not None:
            for key in errors:
                if key.startswith('object__name__'):
                    errors[key] += _(' Batch numbering resulted in name: %(name)s', name=f'"{batch_names[0]["en"]}"' if len(batch_names[0]) == 1 else ', '.join([f'"{name}" ({get_translated_text(logic.languages.get_language_by_lang_code(lang_code).names)})' for lang_code, name in batch_names[0].items()]))
        if object_data is not None and not errors and batch_names is not None:
            data_sequence = []
            object_data = typing.cast(typing.Dict[str, typing.Any], object_data)
            for name in batch_names:
                object_data['name']['text'] = name
                data_sequence.append(deepcopy(object_data))
                try:
                    logic.schemas.validate(data_sequence[-1], schema, strict=True, file_names_by_id=actual_file_names_by_id)
                except logic.errors.ValidationError as e:
                    batch_error_message = _(
                        '"%(error)s" for object name %(name)s',
                        error=e.raw_message,
                        name=f'"{name["en"]}"' if len(name) == 1 else ', '.join(
                            [
                                f'"{n}" ({get_translated_text(logic.languages.get_language_by_lang_code(lang_code).names)})'
                                for lang_code, n in name.items()
                            ]
                        )
                    )
                    if 'batch' in errors:
                        errors['batch'] += '\n' + batch_error_message
                    else:
                        errors['batch'] = batch_error_message
        if form_data['action_submit'] == 'inline_edit' and errors:
            return flask.jsonify({
                'errors': errors
            }), 400
        template_arguments.update({
            'errors_by_title': get_errors_by_title(errors, schema)
        })
        if object_data is not None and not errors:
            object_data = typing.cast(typing.Dict[str, typing.Any], object_data)
            for markdown in logic.markdown_to_html.get_markdown_from_object_data(object_data):
                markdown_as_html = logic.markdown_to_html.markdown_to_safe_html(markdown)
                logic.markdown_images.mark_referenced_markdown_images_as_permanent(markdown_as_html)
            referenced_temporary_file_ids = sorted(logic.temporary_files.get_referenced_temporary_file_ids(object_data))
            if object is None:
                copy_permissions_object_id, permissions_for_group_id, permissions_for_project_id = _parse_permissions_ids(form_data)
                read_permissions_to_all_users = form_data.get('all_users_read_permissions') == '1'
                permanent_file_names_by_id = {}
                permanent_file_map = {}
                for ind, file_id in enumerate(referenced_temporary_file_ids):
                    permanent_file_map[file_id] = ind
                    permanent_file_names_by_id[ind] = actual_file_names_by_id[-file_id]
                logic.temporary_files.replace_file_reference_ids(object_data, permanent_file_map)
                if batch_names is not None:
                    objects = create_object_batch(
                        action_id=action.id,
                        data_sequence=data_sequence,
                        user_id=flask_login.current_user.id,
                        copy_permissions_object_id=copy_permissions_object_id,
                        permissions_for_group_id=permissions_for_group_id,
                        permissions_for_project_id=permissions_for_project_id,
                        permissions_for_all_users=Permissions.READ if read_permissions_to_all_users else None,
                        data_validator_arguments={'file_names_by_id': permanent_file_names_by_id},
                        validate_data=False     # validated on data_sequence creation
                    )
                else:
                    objects = [create_object(
                        action_id=action.id,
                        data=object_data,
                        user_id=flask_login.current_user.id,
                        previous_object_id=previous_object.id if previous_object is not None else None,
                        schema=previous_object.schema if previous_object is not None and not should_upgrade_schema else None,
                        copy_permissions_object_id=copy_permissions_object_id,
                        permissions_for_group_id=permissions_for_group_id,
                        permissions_for_project_id=permissions_for_project_id,
                        permissions_for_all_users=Permissions.READ if read_permissions_to_all_users else None,
                        data_validator_arguments={'file_names_by_id': permanent_file_names_by_id}
                    )]
                object_ids = [object.id for object in objects]
                if action.instrument_id and not flask.current_app.config['DISABLE_INSTRUMENTS'] and may_create_log_entry:
                    if 'create_instrument_log_entry' in form_data:
                        category_ids = []
                        for category_id in raw_form_data.get('instrument_log_categories', []):
                            try:
                                if int(category_id) in [category.id for category in instrument_log_categories]:
                                    category_ids.append(int(category_id))
                            except Exception:
                                pass
                        log_entry = logic.instrument_log_entries.create_instrument_log_entry(
                            instrument_id=action.instrument_id,
                            user_id=flask_login.current_user.id,
                            content='',
                            category_ids=category_ids
                        )
                        for object_id in object_ids:
                            logic.instrument_log_entries.create_instrument_log_object_attachment(
                                instrument_log_entry_id=log_entry.id,
                                object_id=object_id
                            )
                logic.temporary_files.copy_temporary_files(file_ids=referenced_temporary_file_ids, context_id=context_id, user_id=flask_login.current_user.id, object_ids=object_ids)
                logic.temporary_files.delete_temporary_files(context_id=context_id)
                if len(object_ids) == 1:
                    flask.flash(_('The object was created successfully.'), 'success')
                    return flask.redirect(flask.url_for('.object', object_id=object_ids[0]))
                else:
                    flask.flash(_('The objects were created successfully.'), 'success')
                    return flask.redirect(flask.url_for('.objects', ids=','.join([str(object_id) for object_id in object_ids])))
            else:
                if object_data != object.data or schema != object.schema:
                    actual_temporary_file_id_map = logic.temporary_files.copy_temporary_files(file_ids=referenced_temporary_file_ids, context_id=context_id, user_id=flask_login.current_user.id, object_ids=[object.id])
                    logic.temporary_files.delete_temporary_files(context_id=context_id)
                    logic.temporary_files.replace_file_reference_ids(object_data, actual_temporary_file_id_map)
                    update_object(object_id=object.id, user_id=flask_login.current_user.id, data=object_data, schema=schema)
                    flask.flash(_('The object was updated successfully.'), 'success')
                return flask.redirect(flask.url_for('.object', object_id=object.id))

    _update_recipes_for_input(schema)

    if object is None:
        # alternatives to default permissions
        user_groups = logic.groups.get_user_groups(flask_login.current_user.id)
        user_projects = logic.projects.get_user_projects(flask_login.current_user.id, include_groups=True)
        template_arguments.update({
            'can_copy_permissions': True,
            'user_groups': user_groups,
            'user_projects': user_projects
        })

        return flask.render_template(
            'objects/forms/form_create.html',
            action_id=action.id,
            action=action,
            possible_properties=possible_object_id_properties,
            passed_object_ids=passed_object_ids,
            show_selecting_modal=show_selecting_modal,
            **template_arguments
        )
    else:
        return flask.render_template(
            'objects/forms/form_edit.html',
            object_id=object.object_id,
            mode=mode,
            **template_arguments
        )


def _apply_placeholder_data(
        placeholder_data: typing.Dict[typing.Sequence[typing.Union[str, int]], typing.Any],
        data: typing.Dict[str, typing.Any]
) -> None:
    for path, value in placeholder_data.items():
        try:
            sub_data: typing.Optional[typing.Union[typing.List[typing.Any], typing.Dict[str, typing.Any]]] = data
            for step in path[:-1]:
                if isinstance(sub_data, list) and isinstance(step, int):
                    sub_data = sub_data[step]
                elif isinstance(sub_data, dict) and isinstance(step, str):
                    sub_data = sub_data[step]
                else:
                    sub_data = None
                    break
            step = path[-1]
            if isinstance(sub_data, list) and isinstance(step, int):
                sub_data[step] = value
            elif isinstance(sub_data, dict) and isinstance(step, str):
                sub_data[step] = value
        except Exception:
            # Ignore invalid placeholder data
            pass


def _parse_permissions_ids(
        form_data: typing.Dict[str, typing.Any]
) -> typing.Tuple[typing.Optional[int], typing.Optional[int], typing.Optional[int]]:
    """
    Parse and return the IDs for assigning permissions.

    :param form_data: the filtered form data
    :return: the object, basic group and project group IDs as a tuple
    """
    if form_data.get('permissions_method') == 'copy_permissions':
        copy_permissions_object_id_str = form_data.get('copy_permissions_object_id')
        if copy_permissions_object_id_str:
            try:
                copy_permissions_object_id = int(copy_permissions_object_id_str)
                if Permissions.READ in get_user_object_permissions(copy_permissions_object_id, flask_login.current_user.id):
                    return copy_permissions_object_id, None, None
            except Exception:
                pass
            flask.flash(_("Unable to copy permissions. Default permissions will be applied."), 'error')
        else:
            flask.flash(_("No object selected. Default permissions will be applied."), 'error')
        return None, None, None
    if form_data.get('permissions_method') == 'permissions_for_group':
        permissions_for_group_id_str = form_data.get('permissions_for_group_group_id')
        if permissions_for_group_id_str:
            try:
                permissions_for_group_id = int(permissions_for_group_id_str)
                if flask_login.current_user.id in logic.groups.get_group_member_ids(permissions_for_group_id):
                    return None, permissions_for_group_id, None
            except Exception:
                pass
            flask.flash(_("Unable to grant permissions to basic group. Default permissions will be applied."), 'error')
        else:
            flask.flash(_("No basic group selected. Default permissions will be applied."), 'error')
        return None, None, None
    if form_data.get('permissions_method') == 'permissions_for_project':
        permissions_for_project_id_str = form_data.get('permissions_for_project_project_id')
        if permissions_for_project_id_str:
            try:
                permissions_for_project_id = int(permissions_for_project_id_str)
                if flask_login.current_user.id in logic.projects.get_project_member_user_ids_and_permissions(permissions_for_project_id, include_groups=True):
                    return None, None, permissions_for_project_id
            except Exception:
                pass
            flask.flash(_("Unable to grant permissions to project group. Default permissions will be applied."), 'error')
        else:
            flask.flash(_("No project group selected. Default permissions will be applied."), 'error')
        return None, None, None
    return None, None, None


def _handle_batch_names(
        schema: typing.Dict[str, typing.Any],
        form_data: typing.Dict[str, typing.Any],
        raw_form_data: typing.Dict[str, typing.List[typing.Any]],
        errors: typing.Dict[str, str]
) -> typing.Optional[typing.List[typing.Dict[str, str]]]:
    """
    Generate object names for a batch.

    :param schema: the object schema
    :param form_data: the filtered form data, to read values from
    :param raw_form_data: the raw form data, to write the example name into
    :param errors: an error dict, to write errors to
    :return: the batch names as a list of multi-language dicts, or None
    """
    if not schema.get('batch', False) or 'input_num_batch_objects' not in form_data:
        return None

    # parse the number of objects in the batch
    try:
        # the form allows notations like '1.2e1' for '12', however Python can only parse these as floats
        num_objects_in_batch_float = float(form_data['input_num_batch_objects'])
        if num_objects_in_batch_float == int(num_objects_in_batch_float):
            num_objects_in_batch = int(num_objects_in_batch_float)
        else:
            num_objects_in_batch = None
    except ValueError:
        num_objects_in_batch = None
    if num_objects_in_batch is not None:
        form_data['input_num_batch_objects'] = str(num_objects_in_batch)
    if num_objects_in_batch is None or num_objects_in_batch <= 0:
        errors['input_num_batch_objects'] = _('The number of objects in batch must be an positive integer.')
        return None
    if num_objects_in_batch > flask.current_app.config['MAX_BATCH_SIZE']:
        errors['input_num_batch_objects'] = _('The maximum number of objects in one batch is %(max_batch_size)s.', max_batch_size=flask.current_app.config['MAX_BATCH_SIZE'])
        return None
    # parse the first number
    try:
        # the form allows notations like '1.2e1' for '12', however Python can only parse these as floats
        first_number_float = float(form_data['input_first_number_batch'])
        if first_number_float == int(first_number_float):
            first_number = int(first_number_float)
        else:
            first_number = None
    except ValueError:
        first_number = None
    if first_number is None:
        errors['input_first_number_batch'] = _('The first number must be an integer.')
        return None
    form_data['input_first_number_batch'] = str(first_number)

    # The object name might need the batch number to match the pattern
    name_suffix_format = schema.get('batch_name_format', '{:d}')
    try:
        example_name_suffix = name_suffix_format.format(first_number)
    except (ValueError, KeyError):
        name_suffix_format = '{:d}'
        example_name_suffix = ''
    if 'object__name__text' in form_data:
        batch_base_name = {'en': form_data['object__name__text']}
        raw_form_data['object__name__text'] = [batch_base_name['en'] + example_name_suffix]
    else:
        enabled_languages_raw: typing.Union[str, typing.List[str]] = form_data.get('object__name__text_languages', [])
        if isinstance(enabled_languages_raw, str):
            enabled_languages = [enabled_languages_raw]
        else:
            enabled_languages = enabled_languages_raw
        if 'en' not in enabled_languages:
            enabled_languages.append('en')
        batch_base_name = {}
        for language_code in enabled_languages:
            batch_base_name[language_code] = form_data.get('object__name__text_' + language_code, '')
            raw_form_data['object__name__text_' + language_code] = [batch_base_name[language_code] + example_name_suffix]
    return [
        {
            language_code: batch_base_name_str + (name_suffix_format.format(i) if name_suffix_format else '')
            for language_code, batch_base_name_str in batch_base_name.items()
        }
        for i in range(first_number, first_number + num_objects_in_batch)
    ]


def _update_recipes_for_input(schema: typing.Dict[str, typing.Any]) -> None:
    """
    Update a schema so that recipes contain the values and types used for input instead of the internal representation.

    :param schema: the schema to update
    """
    if schema['type'] == 'object':
        if 'recipes' in schema:
            for recipe in schema['recipes']:
                recipe['name'] = get_translated_text(recipe['name'])
                for property_name in recipe['property_values']:
                    property_schema = schema['properties'][property_name]
                    property_value = recipe['property_values'][property_name]
                    if property_schema['type'] == 'datetime':
                        if property_value:
                            value = default_format_datetime(property_value['utc_datetime'])
                        else:
                            value = None
                        property_value = {
                            'value': value,
                            'type': 'datetime'
                        }
                    elif property_schema['type'] == 'quantity':
                        if property_value:
                            units = property_value['units']
                            if units in ['min', 'h']:
                                value = format_time(property_value['magnitude_in_base_units'], units)
                            else:
                                value = custom_format_number(property_value['magnitude'])
                        else:
                            units = None
                            value = None
                        property_value = {
                            'value': value,
                            'units': units,
                            'type': 'quantity'
                        }
                    elif property_schema['type'] == 'text':
                        if property_value:
                            value = property_value['text']
                        else:
                            value = None
                        if 'choices' in property_schema:
                            property_value = {
                                'value': value,
                                'type': 'choice'
                            }
                        elif 'markdown' in property_schema and property_schema['markdown']:
                            property_value = {
                                'value': value,
                                'type': 'markdown'
                            }
                        elif 'multiline' in property_schema and property_schema['multiline']:
                            property_value = {
                                'value': value,
                                'type': 'multiline'
                            }
                        else:
                            property_value = {
                                'value': value,
                                'type': 'text'
                            }
                    elif property_schema['type'] == 'bool':
                        property_value = {
                            'value': property_value['value'],
                            'type': 'bool'
                        }
                    property_value['title'] = get_translated_text(property_schema['title'])
                    recipe['property_values'][property_name] = property_value
        for property_schema in schema['properties'].values():
            _update_recipes_for_input(property_schema)
    elif schema['type'] == 'array':
        _update_recipes_for_input(schema['items'])


def _get_sub_data_and_schema(data: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], id_prefix: str) -> typing.Tuple[typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]], typing.Dict[str, typing.Any]]:
    sub_data: typing.Any
    sub_data = data
    sub_schema = schema
    try:
        key: typing.Union[str, int]
        for key in id_prefix.split('__'):
            if sub_schema['type'] == 'array':
                key = int(key)
                sub_schema = sub_schema['items']
            elif sub_schema['type'] == 'object':
                sub_schema = sub_schema['properties'][key]
            else:
                raise ValueError('invalid type')
            if isinstance(key, int):
                while key >= len(sub_data):
                    sub_data.append(generate_placeholder(sub_schema))
            elif key not in sub_data:
                sub_data[key] = generate_placeholder(sub_schema)
            sub_data = sub_data[key]
        if sub_schema['type'] != 'array':
            raise ValueError('invalid type')
    except (ValueError, KeyError, IndexError, TypeError) as exc:
        raise ValueError('invalid action') from exc
    return sub_data, sub_schema


def _apply_action_to_form_data(action: str, form_data: typing.Dict[str, typing.Any]) -> typing.Dict[str, typing.Any]:
    new_form_data = form_data
    action_id_prefix, action_index_str, action_type = action[len('action_'):].rsplit('__', 2)
    if action_type == 'delete':
        deleted_item_index = int(action_index_str)
        parent_id_prefix = action_id_prefix
        new_form_data = {}
        for name in form_data:
            if not name.startswith(parent_id_prefix + '__') or '__' not in name[len(parent_id_prefix) + 2:]:
                new_form_data[name] = form_data[name]
            else:
                item_index_str, id_suffix = name[len(parent_id_prefix) + 2:].split('__', 1)
                item_index = int(item_index_str)
                if item_index < deleted_item_index:
                    new_form_data[name] = form_data[name]
                if item_index > deleted_item_index:
                    new_name = parent_id_prefix + '__' + str(item_index - 1) + '__' + id_suffix
                    new_form_data[new_name] = form_data[name]
    return new_form_data


def get_object_form_template_kwargs(object_id: typing.Optional[int]) -> typing.Dict[str, typing.Any]:
    template_kwargs = {
        'datetime': datetime,
        'languages': get_languages(),
    }

    # actions and action types
    sorted_actions = get_sorted_actions_for_user(
        user_id=flask_login.current_user.id
    )
    fed_action_type_id_map = {
        action_type.id: action_type.fed_id
        for action_type in logic.action_types.get_action_types()
        if action_type.fed_id is not None and action_type.fed_id < 0
    }
    action_type_id_by_action_id = {
        action_id: fed_action_type_id_map.get(action_type_id, action_type_id) if action_type_id is not None else None
        for action_id, action_type_id in logic.actions.get_action_type_ids_for_action_ids(None).items()
    }
    template_kwargs.update({
        'sorted_actions': sorted_actions,
        'action_type_id_by_action_id': action_type_id_by_action_id,
        'ActionType': models.ActionType,
    })

    # temporary file upload for file fields
    context_id_serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='temporary-file-upload')
    if 'context_id_token' in flask.request.form:
        context_id_token = flask.request.form.get('context_id_token', '')
        try:
            user_id, context_id = context_id_serializer.loads(context_id_token, max_age=flask.current_app.config['TEMPORARY_FILE_TIME_LIMIT'])
        except itsdangerous.BadSignature:
            return flask.abort(400)
        if user_id != flask_login.current_user.id:
            return flask.abort(400)
    else:
        context_id = secrets.token_hex(32)
        context_id_token = context_id_serializer.dumps((flask_login.current_user.id, context_id))

    if object_id is not None:
        file_names_by_id = logic.files.get_file_names_by_id_for_object(object_id)
    else:
        file_names_by_id = {}
    temporary_files = logic.temporary_files.get_files_for_context_id(context_id=context_id)
    for temporary_file in temporary_files:
        file_names_by_id[-temporary_file.id] = temporary_file.file_name, temporary_file.file_name

    template_kwargs.update({
        'context_id': context_id,
        'context_id_token': context_id_token,
        'file_names_by_id': file_names_by_id,
    })

    # previously used tags
    tags = [
        {
            'name': tag.name,
            'uses': tag.uses
        }
        for tag in logic.tags.get_tags()
    ]
    template_kwargs.update({
        'tags': tags,
    })

    # users
    users = get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN'])
    users.sort(key=lambda user: user.id)
    template_kwargs.update({
        'users': users,
    })

    return template_kwargs


def _get_title_by_property_path(
        schema: typing.Any,
        property_path: typing.Sequence[str],
        *,
        index: typing.Optional[str] = None,
        is_root_object: bool = True
) -> str:
    if not isinstance(schema, dict):
        return ''
    schema_title = get_translated_text(schema.get('title', ''))
    if index:
        schema_title += ' #' + index
    if len(property_path) == 0 or not schema_title:
        return schema_title
    if schema.get('type') == 'array':
        return ('' if is_root_object else f'{schema_title} ➜ ') + _get_title_by_property_path(schema.get('items'), property_path[1:], is_root_object=False, index=property_path[0])
    if schema.get('type') == 'object':
        property_schemas = schema.get('properties')
        if isinstance(property_schemas, dict):
            return ('' if is_root_object else f'{schema_title} ➜ ') + _get_title_by_property_path(property_schemas.get(property_path[0]), property_path[1:], is_root_object=False)
    return ''


def get_errors_by_title(
        errors: typing.Dict[str, str],
        schema: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Set[str]]:
    # filter out errors for required fields or None array entries with existing error messages
    missing_required_fields = {}
    invalid_type_entries = {}
    for name, message in errors.items():
        if name.startswith('object__') and name.endswith('__hidden') and message.startswith('missing required property "'):
            parent_property_path = name.split('__')[1:-1]
            property_name = message.split('"')[1]
            missing_required_fields[tuple(parent_property_path + [property_name])] = (name, message)
        if name.startswith('object__') and name.endswith('__hidden') and message.startswith('invalid type (at '):
            parent_property_path = name.split('__')[1:-1]
            property_name = message.split('(at ')[1].split(')')[0]
            invalid_type_entries[tuple(parent_property_path + [property_name])] = (name, message)
    ignorable_errors = set()
    for name, message in errors.items():
        if name.startswith('object__'):
            property_path = tuple(name.split('__')[1:-1])
            for potential_ignorable_errors in (missing_required_fields, invalid_type_entries):
                if tuple(property_path) in potential_ignorable_errors:
                    ignorable_errors.add(potential_ignorable_errors[property_path])

    # construct title -> error messages dict
    errors_by_title: typing.Dict[str, typing.Set[str]] = {}
    for name, message in errors.items():
        if (name, message) in ignorable_errors:
            continue
        title = ''
        if name.startswith('object__'):
            property_path = tuple(name.split('__')[1:-1])
            if property_path:
                title = _get_title_by_property_path(schema, property_path)
        elif name == 'batch':
            title = _('Batch Object Creation')
        if not title:
            title = _('Unknown (%(name)s)', name=name)
        if title in errors_by_title:
            errors_by_title[title].update(message.splitlines())
        else:
            errors_by_title[title] = set(message.splitlines())
    return errors_by_title



# coding: utf-8
"""

"""

import enum
import datetime
import typing

from sqlalchemy.orm import Mapped, Query, relationship

from .. import db
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


@enum.unique
class UserLogEntryType(enum.Enum):
    OTHER = 0
    CREATE_OBJECT = 1
    EDIT_OBJECT = 2
    EDIT_OBJECT_PERMISSIONS = 3
    REGISTER_USER = 4
    INVITE_USER = 5
    EDIT_USER_PREFERENCES = 6
    RESTORE_OBJECT_VERSION = 7
    POST_COMMENT = 8
    UPLOAD_FILE = 9
    CREATE_BATCH = 10
    ASSIGN_LOCATION = 11
    CREATE_LOCATION = 12
    UPDATE_LOCATION = 13
    LINK_PUBLICATION = 14
    CREATE_INSTRUMENT_LOG_ENTRY = 15
    IMPORT_FROM_ELN_FILE = 16


class UserLogEntry(Model):
    __tablename__ = 'user_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[UserLogEntryType] = db.Column(db.Enum(UserLogEntryType), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    user: Mapped['User'] = relationship('User', back_populates="log_entries")

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["UserLogEntry"]]

    def __init__(
            self,
            type: UserLogEntryType,
            user_id: int,
            data: typing.Dict[str, typing.Any],
            utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            type=type,
            user_id=user_id,
            data=data,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, user_id={self.user_id}, utc_datetime={self.utc_datetime}, data={self.data})>'



# coding: utf-8
"""
Status route to allow health checks.
"""

import flask


from . import frontend
from .. import db
from .. import version
from ..utils import FlaskResponseT


@frontend.route('/status/')
def status() -> FlaskResponseT:
    status_info = {
        'sampledb_version': version.__version__,
        'flask_status': True
    }
    try:
        db.session.execute(db.text('select 1')).fetchone()
        status_info['postgres_status'] = True
    except Exception:
        status_info['postgres_status'] = False
    if all(status_info.values()):
        status_code = 200
    else:
        status_code = 500
    status_info['name'] = flask.current_app.config['SERVICE_NAME']
    status_info['federation_uuid'] = flask.current_app.config['FEDERATION_UUID'] if flask.current_app.config['ENABLE_FEDERATION_DISCOVERABILITY'] else None
    return flask.jsonify(status_info), status_code, {
        'Access-Control-Allow-Origin': '*'
    }



# coding: utf-8
"""
Replace the default_public_permissions table with the all_user_default_permissions table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    table_exists = db.session.execute(db.text("""
        SELECT table_name
        FROM information_schema.tables
        WHERE table_name = 'default_public_permissions'
    """)).fetchall()
    if not table_exists:
        return False

    # Perform migration
    default_public_permissions = db.session.execute(db.text("""
        SELECT creator_id
        FROM default_public_permissions
        WHERE is_public = true
    """)).fetchall()
    all_user_default_permissions = db.session.execute(db.text("""
        SELECT creator_id, permissions
        FROM all_user_default_permissions
    """)).fetchall()
    all_user_object_permissions = {
        creator_id: permissions
        for creator_id, permissions in all_user_default_permissions
    }
    for creator_id, in default_public_permissions:
        if creator_id not in all_user_object_permissions:
            db.session.execute(db.text("""
                INSERT INTO all_user_default_permissions
                (creator_id, permissions)
                VALUES (:creator_id, 'READ')
            """), {'creator_id': creator_id})
    db.session.execute(db.text("""
        DROP TABLE default_public_permissions
    """))
    return True



# coding: utf-8
"""

"""
import dataclasses
import typing

import flask
import sqlalchemy
from sqlalchemy.dialects import postgresql

from .. import db
from . import errors
from . import actions
from . import action_types
from . import components
from .default_permissions import get_default_permissions_for_users, get_default_permissions_for_groups, get_default_permissions_for_projects, get_default_permissions_for_all_users
from . import instruments
from .notifications import create_notification_for_having_received_an_objects_permissions_request
from . import objects
from ..models import Permissions, UserObjectPermissions, GroupObjectPermissions, ProjectObjectPermissions, AllUserObjectPermissions, AnonymousUserObjectPermissions, Action, Object, Objects
from .users import get_user
from .permissions import ResourcePermissions


__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


object_permissions = ResourcePermissions(
    resource_id_name='object_id',
    all_user_permissions_table=AllUserObjectPermissions,
    anonymous_user_permissions_table=AnonymousUserObjectPermissions,
    user_permissions_table=UserObjectPermissions,
    group_permissions_table=GroupObjectPermissions,
    project_permissions_table=ProjectObjectPermissions,
    check_resource_exists=lambda resource_id: objects.check_object_exists(object_id=resource_id)
)


def get_object_permissions_for_all_users(object_id: int) -> Permissions:
    return object_permissions.get_permissions_for_all_users(resource_id=object_id)


def set_object_permissions_for_all_users(object_id: int, permissions: Permissions) -> None:
    object_permissions.set_permissions_for_all_users(resource_id=object_id, permissions=permissions)


def get_object_permissions_for_anonymous_users(object_id: int) -> Permissions:
    return object_permissions.get_permissions_for_anonymous_users(resource_id=object_id)


def set_object_permissions_for_anonymous_users(object_id: int, permissions: Permissions) -> None:
    object_permissions.set_permissions_for_anonymous_users(resource_id=object_id, permissions=permissions)


def get_object_permissions_for_users(
        object_id: int,
        include_instrument_responsible_users: bool = True,
        include_groups: bool = True,
        include_projects: bool = True,
        include_readonly: bool = True,
        include_admin_permissions: bool = True
) -> typing.Dict[int, Permissions]:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        include_instrument_responsible_users = False
    additional_permissions = {}
    if include_instrument_responsible_users:
        for user_id in _get_object_responsible_user_ids(object_id):
            additional_permissions[user_id] = Permissions.GRANT
    return object_permissions.get_permissions_for_users(
        resource_id=object_id,
        include_groups=include_groups,
        include_projects=include_projects,
        include_admin_permissions=include_admin_permissions,
        limit_readonly_users=include_readonly,
        additional_permissions=additional_permissions
    )


def set_user_object_permissions(object_id: int, user_id: int, permissions: Permissions) -> None:
    object_permissions.set_permissions_for_user(resource_id=object_id, user_id=user_id, permissions=permissions)


def get_object_permissions_for_groups(object_id: int, include_projects: bool = False) -> typing.Dict[int, Permissions]:
    return object_permissions.get_permissions_for_groups(resource_id=object_id, include_projects=include_projects)


def set_group_object_permissions(object_id: int, group_id: int, permissions: Permissions) -> None:
    object_permissions.set_permissions_for_group(resource_id=object_id, group_id=group_id, permissions=permissions)


def get_object_permissions_for_projects(object_id: int) -> typing.Dict[int, Permissions]:
    return object_permissions.get_permissions_for_projects(resource_id=object_id)


def set_project_object_permissions(object_id: int, project_id: int, permissions: Permissions) -> None:
    object_permissions.set_permissions_for_project(resource_id=object_id, project_id=project_id, permissions=permissions)


def _get_object_responsible_user_ids(object_id: int) -> typing.List[int]:
    object = objects.get_object(object_id)
    if object.action_id is None:
        return []
    try:
        action = actions.get_action(object.action_id)
    except errors.ActionDoesNotExistError:
        return []
    if action.instrument_id is None:
        return []
    instrument = instruments.get_instrument(action.instrument_id)
    return [user.id for user in instrument.responsible_users]


def get_user_object_permissions(
        object_id: int,
        user_id: typing.Optional[int],
        include_instrument_responsible_users: bool = True,
        include_groups: bool = True,
        include_projects: bool = True,
        include_readonly: bool = True,
        include_admin_permissions: bool = True
) -> Permissions:
    if flask.current_app.config['DISABLE_INSTRUMENTS']:
        include_instrument_responsible_users = False

    # ensure the object exists to avoid returning permissions for non-existing objects
    objects.check_object_exists(object_id=object_id)

    if user_id is None:
        if flask.current_app.config['ENABLE_ANONYMOUS_USERS']:
            return object_permissions.get_permissions_for_anonymous_users(resource_id=object_id)
        else:
            return Permissions.NONE

    user = get_user(user_id)

    if include_admin_permissions:
        # administrators have GRANT permissions if they use admin permissions
        if user.has_admin_permissions:
            # unless they are limited to READ permissions
            if user.is_readonly:
                return Permissions.READ
            else:
                return Permissions.GRANT

    # select permissions from view for efficiency
    if include_instrument_responsible_users and include_groups and include_projects:
        stmt = db.text("""
        SELECT
        MAX(permissions_int)
        FROM user_object_permissions_by_all
        WHERE (user_id = :user_id OR user_id IS NULL) AND (object_id = :object_id) AND (requires_anonymous_users IS FALSE OR :enable_anonymous_users IS TRUE) AND (requires_instruments IS FALSE OR :enable_instruments IS TRUE)
        """)
        permissions_int_result = db.session.execute(stmt, {
            'user_id': user_id,
            'object_id': object_id,
            'enable_anonymous_users': flask.current_app.config['ENABLE_ANONYMOUS_USERS'],
            'enable_instruments': not flask.current_app.config['DISABLE_INSTRUMENTS']
        }).fetchone()
        if permissions_int_result is None:
            return Permissions.NONE
        permissions_int = permissions_int_result[0]
        if permissions_int is None or permissions_int <= 0:
            return Permissions.NONE
        elif include_readonly and user.is_readonly and permissions_int in (1, 2, 3):
            return Permissions.READ
        elif permissions_int == 1:
            return Permissions.READ
        elif permissions_int == 2:
            return Permissions.WRITE
        elif permissions_int >= 3:
            return Permissions.GRANT

    additional_permissions = Permissions.NONE
    if include_instrument_responsible_users:
        # instrument responsible users have GRANT permissions for an object created with instrument
        if user_id in _get_object_responsible_user_ids(object_id):
            additional_permissions = max(additional_permissions, Permissions.GRANT)

    return object_permissions.get_permissions_for_user(
        resource_id=object_id,
        user_id=user_id,
        include_all_users=True,
        # user is not anonymous but may still benefit from anonymous user permissions
        include_anonymous_users=flask.current_app.config['ENABLE_ANONYMOUS_USERS'],
        include_groups=include_groups,
        include_projects=include_projects,
        # admin permissions have already been checked above
        include_admin_permissions=False,
        limit_readonly_users=include_readonly,
        additional_permissions=additional_permissions
    )


def set_initial_permissions(obj: Object, user_id: typing.Optional[int] = None) -> None:
    if user_id is None:
        user_id = obj.user_id
    if user_id is None:
        # no global default permissions
        return
    default_user_permissions = get_default_permissions_for_users(creator_id=user_id)
    for other_user_id, permissions in default_user_permissions.items():
        set_user_object_permissions(object_id=obj.object_id, user_id=other_user_id, permissions=permissions)
    default_group_permissions = get_default_permissions_for_groups(creator_id=user_id)
    for group_id, permissions in default_group_permissions.items():
        set_group_object_permissions(object_id=obj.object_id, group_id=group_id, permissions=permissions)
    default_project_permissions = get_default_permissions_for_projects(creator_id=user_id)
    for project_id, permissions in default_project_permissions.items():
        set_project_object_permissions(object_id=obj.object_id, project_id=project_id, permissions=permissions)
    permissions_for_all_users = get_default_permissions_for_all_users(creator_id=user_id)
    set_object_permissions_for_all_users(object_id=obj.object_id, permissions=permissions_for_all_users)


@dataclasses.dataclass(frozen=True)
class ObjectInfo:
    object_id: int
    name_json: typing.Optional[typing.Union[str, typing.Dict[str, str]]]
    action_id: typing.Optional[int]
    max_permission: int
    tags: typing.Optional[typing.Dict[str, typing.Any]]
    fed_object_id: typing.Optional[int]
    component_name: typing.Optional[str]
    eln_object_id: typing.Optional[str]
    eln_import_id: typing.Optional[int]


def get_object_info_with_permissions(
        user_id: int,
        permissions: Permissions,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        action_id: typing.Optional[int] = None,
        action_ids: typing.Optional[typing.Sequence[int]] = None,
        action_type_id: typing.Optional[int] = None,
        object_ids: typing.Optional[typing.Sequence[int]] = None
) -> typing.List[ObjectInfo]:

    user = get_user(user_id)

    # readonly users may not have more than READ permissions
    if user.is_readonly and permissions != Permissions.READ:
        return []

    action_filters = []
    if action_type_id is not None:
        action_filters.append(Action.type_id == action_type_id)
    if action_id is not None:
        action_filters.append(Action.id == action_id)
    if action_ids is not None:
        action_filters.append(Action.id.in_(tuple(action_ids)))
    if action_filters:
        action_filter = action_filters.pop(0)
        for additional_action_filter in action_filters:
            action_filter = db.and_(action_filter, additional_action_filter)
    else:
        action_filter = None

    if user.has_admin_permissions:
        # admins who use admin permissions do not need permission-based filtering
        stmt = db.text("""
        SELECT
            o.object_id, o.name_cache AS name_json, o.action_id, 3 AS max_permission, o.tags_cache as tags, o.fed_object_id, COALESCE(c.name, c.address, c.uuid) AS component_name, o.eln_import_id, o.eln_object_id
        FROM objects_current AS o
        LEFT JOIN components AS c ON c.id = o.component_id
        """)
    else:
        stmt = db.text("""
        SELECT
            o.object_id, o.name_cache AS name_json, o.action_id, p.max_permission, o.tags_cache as tags, o.fed_object_id, COALESCE(c.name, c.address, c.uuid) AS component_name, o.eln_import_id, o.eln_object_id
        FROM (
            SELECT
            object_id, MAX(permissions_int) AS max_permission
            FROM user_object_permissions_by_all
            WHERE (user_id = :user_id OR user_id IS NULL) AND (requires_anonymous_users IS FALSE OR :enable_anonymous_users IS TRUE) AND (requires_instruments IS FALSE OR :enable_instruments IS TRUE)
            GROUP BY (object_id)
            HAVING MAX(permissions_int) >= :min_permissions_int
        ) AS p
        JOIN objects_current AS o ON o.object_id = p.object_id
        LEFT JOIN components AS c ON c.id = o.component_id
        """)

    stmt = stmt.columns(
        Objects._current_table.c.object_id,
        sqlalchemy.sql.expression.column('name_json'),
        Objects._current_table.c.action_id,
        sqlalchemy.sql.expression.column('max_permission'),
        sqlalchemy.sql.expression.column('tags'),
        sqlalchemy.sql.expression.column('fed_object_id'),
        sqlalchemy.sql.expression.column('component_name'),
        sqlalchemy.sql.expression.column('eln_import_id'),
        sqlalchemy.sql.expression.column('eln_object_id'),
    )

    parameters = {
        'min_permissions_int': permissions.value,
        'user_id': user_id,
        'enable_anonymous_users': flask.current_app.config['ENABLE_ANONYMOUS_USERS'],
        'enable_instruments': not flask.current_app.config['DISABLE_INSTRUMENTS']
    }

    table = stmt.alias('tbl')

    if action_filter is not None:
        where = table.c.action_id.in_(db.select(Action.__table__.c.id).where(action_filter))
    else:
        where = None

    if object_ids is not None:
        object_id_where = table.c.object_id.in_(tuple(object_ids))
        if where is None:
            where = object_id_where
        else:
            where = db.and_(where, object_id_where)

    stmt = table.select()
    if where is not None:
        stmt = stmt.where(where)
    stmt = stmt.order_by(db.desc(table.c.object_id)).limit(limit).offset(offset)

    object_infos = db.session.execute(stmt, parameters).fetchall()

    return [
        ObjectInfo(
            object_id=object_info.object_id,
            name_json=object_info.name_json,
            action_id=object_info.action_id,
            max_permission=object_info.max_permission,
            tags=object_info.tags,
            fed_object_id=object_info.fed_object_id,
            component_name=object_info.component_name,
            eln_object_id=object_info.eln_object_id,
            eln_import_id=object_info.eln_import_id
        )
        for object_info in object_infos
    ]


def get_user_permissions_for_multiple_objects(
        user_id: typing.Optional[int],
        object_ids: typing.Sequence[int]
) -> typing.Dict[int, Permissions]:
    if not object_ids:
        return {}
    if user_id is None:
        if not flask.current_app.config['ENABLE_ANONYMOUS_USERS']:
            return {
                object_id: Permissions.NONE
                for object_id in object_ids
            }
    else:
        user = get_user(user_id)
        if user.has_admin_permissions:
            return {
                object_id: Permissions.GRANT
                for object_id in object_ids
            }

    stmt = db.text("""
    SELECT
    u.object_id, MAX(u.permissions_int)
    FROM user_object_permissions_by_all as u
    WHERE (u.object_id IN :object_ids) AND (u.user_id = :user_id OR u.user_id IS NULL) AND (u.requires_anonymous_users IS FALSE OR :enable_anonymous_users IS TRUE) AND (u.requires_instruments IS FALSE OR :enable_instruments IS TRUE)
    GROUP BY (u.object_id)
    """)

    parameters: typing.Dict[str, typing.Any] = {
        'user_id': user_id,
        'enable_anonymous_users': flask.current_app.config['ENABLE_ANONYMOUS_USERS'],
        'enable_instruments': not flask.current_app.config['DISABLE_INSTRUMENTS'],
        'object_ids': tuple(object_ids),
    }
    result = {
        object_id: Permissions.from_value(permissions_int)
        for object_id, permissions_int in db.session.execute(stmt, parameters).fetchall()
    }
    # use NONE for all missing objects as fallback
    for object_id in object_ids:
        if object_id not in result:
            result[object_id] = Permissions.NONE
    return result


def get_object_table_with_permissions(
        user_id: typing.Optional[int],
        permissions: Permissions,
        other_user_id: typing.Optional[int] = None,
        other_user_permissions: typing.Optional[Permissions] = None,
        group_id: typing.Optional[int] = None,
        group_permissions: typing.Optional[Permissions] = None,
        project_id: typing.Optional[int] = None,
        project_permissions: typing.Optional[Permissions] = None,
        all_users_permissions: typing.Optional[Permissions] = None,
        anonymous_users_permissions: typing.Optional[Permissions] = None,
        object_ids: typing.Optional[typing.Sequence[int]] = None,
        name_only: bool = False,
) -> typing.Union[typing.Tuple[None, None], typing.Tuple[typing.Any, typing.Dict[str, typing.Any]]]:
    if user_id is None:
        if not flask.current_app.config['ENABLE_ANONYMOUS_USERS']:
            return None, None
        user = None
    else:
        user = get_user(user_id)

        # readonly users may not have more than READ permissions
        if user.is_readonly and permissions != Permissions.READ:
            return None, None

    parameters: typing.Dict[str, typing.Any] = {
        'min_permissions_int': permissions.value,
        'user_id': user_id
    }

    if name_only:
        stmt = """
        SELECT
        o.object_id, o.version_id, o.action_id, jsonb_set('{"name": {"_type": "text", "text": ""}}', '{name,text}', o.name_cache::jsonb) as data, '{"title": "Object", "type": "object", "properties": {"name": {"title": "Name", "type": "text"}}}'::jsonb as schema, o.user_id, o.utc_datetime, o.fed_object_id, o.fed_version_id, o.component_id, o.eln_import_id, o.eln_object_id, o.data as data_full
        FROM objects_current AS o
        """
    else:
        stmt = """
        SELECT
        o.object_id, o.version_id, o.action_id, o.data, o.schema, o.user_id, o.utc_datetime, o.fed_object_id, o.fed_version_id, o.component_id, o.eln_import_id, o.eln_object_id, o.data as data_full
        FROM objects_current AS o
        """

    if user is None or not user.has_admin_permissions:
        stmt += """
        JOIN (
            SELECT
            u.object_id
            FROM user_object_permissions_by_all as u
            WHERE (u.user_id = :user_id OR u.user_id IS NULL) AND (u.requires_anonymous_users IS FALSE OR :enable_anonymous_users IS TRUE) AND (u.requires_instruments IS FALSE OR :enable_instruments IS TRUE)
            GROUP BY (u.object_id)
            HAVING MAX(u.permissions_int) >= :min_permissions_int
        ) AS up
        ON o.object_id = up.object_id
        """
        parameters['enable_anonymous_users'] = flask.current_app.config['ENABLE_ANONYMOUS_USERS']
        parameters['enable_instruments'] = not flask.current_app.config['DISABLE_INSTRUMENTS']

    if other_user_id is not None:
        stmt += """
        JOIN (
            SELECT
            u.object_id
            FROM user_object_permissions_by_all as u
            WHERE (u.user_id = :other_user_id OR u.user_id IS NULL) AND (u.requires_anonymous_users IS FALSE OR :enable_anonymous_users IS TRUE) AND (u.requires_instruments IS FALSE OR :enable_instruments IS TRUE)
            GROUP BY (u.object_id)
            HAVING MAX(u.permissions_int) >= :min_other_user_permissions_int
        ) AS oup
        ON o.object_id = oup.object_id
        """
        parameters['other_user_id'] = other_user_id
        if other_user_permissions is None:
            other_user_permissions = permissions
        parameters['min_other_user_permissions_int'] = other_user_permissions.value
        parameters['enable_anonymous_users'] = flask.current_app.config['ENABLE_ANONYMOUS_USERS']
        parameters['enable_instruments'] = not flask.current_app.config['DISABLE_INSTRUMENTS']

    if project_id is not None:
        stmt += """
        JOIN project_object_permissions as pp
        ON (
            pp.object_id = o.object_id AND
            pp.project_id = :project_id AND
            ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> pp.permissions::text)::int >= :min_project_permissions_int
        )
        """
        parameters['project_id'] = project_id
        if project_permissions is None:
            project_permissions = permissions
        parameters['min_project_permissions_int'] = project_permissions.value

    if group_id is not None:
        stmt += """
        JOIN group_object_permissions as gp
        ON (
            gp.object_id = o.object_id AND
            gp.group_id = :group_id AND
            ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> gp.permissions::text)::int >= :min_group_permissions_int
        )
        """
        parameters['group_id'] = group_id
        if group_permissions is None:
            group_permissions = permissions
        parameters['min_group_permissions_int'] = group_permissions.value

    if all_users_permissions is not None:
        if flask.current_app.config['ENABLE_ANONYMOUS_USERS']:
            stmt += """
            JOIN (
                SELECT object_id, permissions FROM all_user_object_permissions
                UNION
                SELECT object_id, permissions FROM anonymous_user_object_permissions
            ) as ap1
            ON (
                ap1.object_id = o.object_id AND
                ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> ap1.permissions::text)::int >= :min_all_users_permissions_int
            )
            """
        else:
            stmt += """
            JOIN all_user_object_permissions as ap1
            ON (
                ap1.object_id = o.object_id AND
                ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> ap1.permissions::text)::int >= :min_all_users_permissions_int
            )
            """
        parameters['min_all_users_permissions_int'] = all_users_permissions.value

    if flask.current_app.config['ENABLE_ANONYMOUS_USERS']:
        if anonymous_users_permissions is not None:
            stmt += """
            JOIN anonymous_user_object_permissions as ap2
            ON (
                ap2.object_id = o.object_id AND
                ('{"READ": 1, "WRITE": 2, "GRANT": 3}'::jsonb ->> ap2.permissions::text)::int >= :min_anonymous_users_permissions_int
            )
            """
            parameters['min_anonymous_users_permissions_int'] = anonymous_users_permissions.value
    elif anonymous_users_permissions is not None and Permissions.READ in anonymous_users_permissions:
        # if permissions for anonymous users are required but anonymous users are disabled, do not return any objects
        return None, None

    if object_ids:
        stmt += """
        WHERE o.object_id IN :object_ids
        """
        parameters['object_ids'] = tuple(object_ids)
    elif object_ids is not None:
        stmt += """
        WHERE FALSE
        """

    table = db.text(stmt).columns(
        Objects._current_table.c.object_id,
        Objects._current_table.c.version_id,
        Objects._current_table.c.action_id,
        Objects._current_table.c.data,
        Objects._current_table.c.schema,
        Objects._current_table.c.user_id,
        Objects._current_table.c.utc_datetime,
        Objects._current_table.c.fed_object_id,
        Objects._current_table.c.fed_version_id,
        Objects._current_table.c.component_id,
        Objects._current_table.c.eln_import_id,
        Objects._current_table.c.eln_object_id,
        db.column('data_full', postgresql.JSONB)
    ).subquery()
    return table, parameters


def get_objects_with_permissions(
        user_id: typing.Optional[int],
        permissions: Permissions,
        filter_func: typing.Callable[[typing.Any], typing.Any] = lambda data: True,
        sorting_func: typing.Optional[typing.Callable[[typing.Any], typing.Any]] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        action_ids: typing.Optional[typing.List[int]] = None,
        action_type_ids: typing.Optional[typing.List[int]] = None,
        other_user_id: typing.Optional[int] = None,
        other_user_permissions: typing.Optional[Permissions] = None,
        group_id: typing.Optional[int] = None,
        group_permissions: typing.Optional[Permissions] = None,
        project_id: typing.Optional[int] = None,
        project_permissions: typing.Optional[Permissions] = None,
        all_users_permissions: typing.Optional[Permissions] = None,
        anonymous_users_permissions: typing.Optional[Permissions] = None,
        object_ids: typing.Optional[typing.Sequence[int]] = None,
        num_objects_found: typing.Optional[typing.List[int]] = None,
        name_only: bool = False,
        **kwargs: typing.Dict[str, typing.Any]
) -> typing.List[Object]:
    if object_ids is not None and not object_ids:
        return []
    table, parameters = get_object_table_with_permissions(
        user_id=user_id,
        permissions=permissions,
        other_user_id=other_user_id,
        other_user_permissions=other_user_permissions,
        group_id=group_id,
        group_permissions=group_permissions,
        project_id=project_id,
        project_permissions=project_permissions,
        all_users_permissions=all_users_permissions,
        anonymous_users_permissions=anonymous_users_permissions,
        object_ids=object_ids,
        name_only=name_only,
    )
    if table is None:
        return []

    if action_type_ids is not None and any(action_type_id <= 0 for action_type_id in action_type_ids):
        # include federated equivalents for default action types
        all_action_types = action_types.get_action_types()
        fed_default_action_types: typing.Dict[int, typing.List[int]] = {}
        for action_type in all_action_types:
            if action_type.fed_id is not None and action_type.fed_id <= 0:
                if action_type.fed_id not in fed_default_action_types:
                    fed_default_action_types[action_type.fed_id] = []
                fed_default_action_types[action_type.fed_id].append(action_type.id)
        extended_action_type_ids = set(action_type_ids)
        for action_type_id in action_type_ids:
            if action_type_id <= 0:
                extended_action_type_ids.update(fed_default_action_types.get(action_type_id, []))
        action_type_ids = list(extended_action_type_ids)

    if action_type_ids is not None and action_ids is not None:
        action_filter = db.and_(Action.type_id.in_(action_type_ids), Action.id.in_(action_ids))
    elif action_type_ids is not None:
        action_filter = Action.type_id.in_(action_type_ids)
    elif action_ids is not None:
        action_filter = Action.id.in_(action_ids)
    else:
        action_filter = None

    return objects.get_objects(
        filter_func=filter_func,
        action_filter=action_filter,
        table=table,
        parameters=parameters,
        sorting_func=sorting_func,
        limit=limit,
        offset=offset,
        num_objects_found=num_objects_found,
        **kwargs
    )


def request_object_permissions(requester_id: int, object_id: int) -> None:
    permissions_by_user = get_object_permissions_for_users(object_id)
    if Permissions.READ in permissions_by_user.get(requester_id, Permissions.NONE):
        return
    granting_user_ids = [
        user_id
        for user_id, permissions in permissions_by_user.items()
        if Permissions.GRANT in permissions
    ]
    for user_id in granting_user_ids:
        create_notification_for_having_received_an_objects_permissions_request(user_id, object_id, requester_id)


def copy_permissions(target_object_id: int, source_object_id: int) -> None:
    object_permissions.copy_permissions(source_resource_id=source_object_id, target_resource_id=target_object_id)


def get_object_if_user_has_permissions(user_id: int, permissions: Permissions, object_id: int, component_uuid: typing.Optional[str] = None) -> typing.Optional[Object]:
    if component_uuid is None or component_uuid == flask.current_app.config['FEDERATION_UUID']:
        try:
            if permissions not in get_user_object_permissions(object_id, user_id):
                return None
            else:
                return objects.get_object(object_id)
        except errors.ObjectDoesNotExistError:
            return None
    else:
        try:
            component = components.get_component_by_uuid(component_uuid)
        except errors.ComponentDoesNotExistError:
            return None
        try:
            object = objects.get_fed_object(object_id, component.id)
            if permissions not in get_user_object_permissions(object.id, user_id):
                return None
        except errors.ObjectDoesNotExistError:
            return None
        return object



# coding: utf-8
"""

"""

import typing

import pint

from ..units import ureg
from ..errors import UndefinedUnitError

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def units_are_valid(units: str) -> bool:
    try:
        ureg.Unit(units)
        return True
    except pint.UndefinedUnitError:
        return False


def get_dimensionality_for_units(
        units: typing.Union[str, typing.List[str]]
) -> str:
    """
    Returns the units' dimensionality in the dimensionality syntax of the pint package.

    :param units: a valid
    :return: dimensionality as string
    :raise errors.UndefinedUnitError: if the units are undefined
    """
    if isinstance(units, list):
        if len(units) > 0:
            units = units[0]
        else:
            units = '1'
    try:
        return str(ureg.Unit(units).dimensionality)
    except pint.UndefinedUnitError:
        raise UndefinedUnitError()


def get_property_paths_for_schema(
        schema: typing.Dict[str, typing.Any],
        valid_property_types: typing.Optional[typing.Set[str]] = None,
        path: typing.Optional[typing.List[typing.Optional[str]]] = None,
        path_depth_limit: typing.Optional[int] = None
) -> typing.Dict[
    typing.Sequence[typing.Optional[str]],
    typing.Dict[str, typing.Optional[typing.Union[str, typing.Dict[str, str]]]]
]:
    """
    Get a dict mapping property paths to the type used in the given schema.

    None is used as a placeholder for array indices and types may be limited
    to a set of valid types.

    :param schema: the schema to generate the dict for
    :param valid_property_types: a set of property types, or None
    :param path: the path to this subschema, or None
    :param path_depth_limit: how deep paths may be, or None
    :return: the generated dict
    """
    return {
        property_path: {
            'type': property_schema.get('type'),
            'title': property_schema.get('title')
        }
        for property_path, property_schema in schema_iter(
            schema=schema,
            path=tuple(path or []),
            filter_property_types=valid_property_types,
            filter_path_depth_limit=path_depth_limit
        )
        if property_schema.get('type') is not None
    }


def schema_iter(
        schema: typing.Dict[str, typing.Any],
        *,
        path: typing.Sequence[typing.Optional[str]] = (),
        filter_path_depth_limit: typing.Optional[int] = None,
        filter_property_types: typing.Optional[typing.Set[str]] = None,
        filter_property_path: typing.Optional[typing.Sequence[typing.Optional[str]]] = None
) -> typing.Iterator[typing.Tuple[typing.Sequence[typing.Optional[str]], typing.Dict[str, typing.Any]]]:
    """
    Iterate over a schema.

    :param schema: the schema to iterate over
    :param path: the path to this schema
    :param filter_path_depth_limit: how deep paths may be, or None
    :param filter_property_types: a set of property types to include in the
        iterator, or None
    :param filter_property_path: a path to return the schema and sub-schemas
        for, or None
    :return: an iterator over tuples of path and schema at that path
    """
    if filter_path_depth_limit is not None and len(path) > filter_path_depth_limit:
        return
    if filter_property_path is not None and path[:min(len(path), len(filter_property_path))] != filter_property_path[:min(len(path), len(filter_property_path))]:
        return
    schema_type = schema.get('type')
    if (
            (filter_property_types is None or schema_type in filter_property_types) and
            (filter_property_path is None or path[:len(filter_property_path)] == filter_property_path)
    ):
        yield path, schema
    if schema_type == 'object':
        schema_properties = schema.get('properties')
        if isinstance(schema_properties, dict):
            for property_name, property_schema in schema_properties.items():
                if isinstance(property_name, str) and isinstance(property_schema, dict):
                    yield from schema_iter(
                        schema=property_schema,
                        path=tuple(path) + (property_name,),
                        filter_path_depth_limit=filter_path_depth_limit,
                        filter_property_types=filter_property_types,
                        filter_property_path=filter_property_path
                    )
    if schema_type == 'array':
        item_schema = schema.get('items')
        if isinstance(item_schema, dict):
            yield from schema_iter(
                schema=item_schema,
                path=tuple(path) + (None,),
                filter_path_depth_limit=filter_path_depth_limit,
                filter_property_types=filter_property_types,
                filter_property_path=filter_property_path
            )


def data_iter(
        data: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]],
        *,
        path: typing.Sequence[typing.Union[str, int]] = (),
        filter_property_types: typing.Optional[typing.Set[str]] = None,
) -> typing.Iterator[typing.Tuple[typing.Sequence[typing.Union[str, int]], typing.Union[typing.Dict[str, typing.Any], typing.Sequence[typing.Any]]]]:
    """
    Iterate over data.

    :param data: the data to iterate over
    :param path: the path to this schema
    :param filter_property_types: a set of property types to include in the
        iterator, or None
    :return: an iterator over tuples of path and data at that path
    """
    if isinstance(data, dict):
        if '_type' in data:
            data_type = data['_type']
        else:
            data_type = 'object'
    elif isinstance(data, list):
        data_type = 'array'
    else:
        return
    if filter_property_types is None or data_type in filter_property_types:
        yield path, data
    if data_type == 'object' and isinstance(data, dict):
        for property_name, property_data in data.items():
            if isinstance(property_name, str) and isinstance(property_data, (dict, list)):
                yield from data_iter(
                    data=property_data,
                    path=tuple(path) + (property_name,),
                    filter_property_types=filter_property_types,
                )
    if data_type == 'array':
        for index, item_data in enumerate(data):
            yield from data_iter(
                data=item_data,
                path=tuple(path) + (index,),
                filter_property_types=filter_property_types,
            )



# coding: utf-8
"""

"""

import datetime
import json
import typing

import itsdangerous
import flask
import flask_login
from flask_babel import _, lazy_gettext, refresh
from fido2.webauthn import UserVerificationRequirement
import werkzeug.wrappers.response

from .. import frontend
from ...logic import oidc
from ...logic.authentication import login, get_active_two_factor_authentication_methods, get_all_fido2_passkey_credentials, get_user_id_for_fido2_passkey_credential_id, get_webauthn_server
from ...logic.users import get_user, User, create_sampledb_federated_identity, get_federated_identity, enable_federated_identity_for_login
from ...logic.components import get_component, get_federated_login_components
from ...logic.errors import FederatedUserInFederatedIdentityError, UserDoesNotExistError, FederatedIdentityNotFoundError
from ..users_forms import SigninForm, SignoutForm
from .forms import WebAuthnLoginForm
from ... import login_manager
from ...utils import FlaskResponseT


@login_manager.user_loader
def load_user(user_id: typing.Any) -> typing.Optional[User]:
    flask.g.shared_device_state = False
    if flask.session.get('using_shared_device') and flask.request.cookies.get('SAMPLEDB_LAST_ACTIVITY_DATETIME'):
        try:
            millisecond_offset = float(flask.request.cookies.get('SAMPLEDB_LAST_ACTIVITY_MILLISECOND_OFFSET', '0'))
        except Exception:
            millisecond_offset = 0
        offset = datetime.timedelta(milliseconds=millisecond_offset)
        try:
            last_activity_datetime = datetime.datetime.strptime(flask.request.cookies['SAMPLEDB_LAST_ACTIVITY_DATETIME'], '%Y-%m-%dT%H:%M:%S')
        except Exception:
            flask.g.shared_device_state = True
        else:
            idle_sign_out_minutes = flask.current_app.config['SHARED_DEVICE_SIGN_OUT_MINUTES']
            idle_minutes = (datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None) + offset - last_activity_datetime).total_seconds() / 60 - 5 / 60
            flask.g.shared_device_state = idle_minutes < idle_sign_out_minutes
            if idle_minutes >= idle_sign_out_minutes:
                flask.session.clear()
                if user_id:
                    flask.flash(_('You have been signed out after %(idle_sign_out_minutes)s minutes of inactivity.', idle_sign_out_minutes=idle_sign_out_minutes), 'info')
                return None
    try:
        user_id = int(user_id)
    except ValueError:
        return None
    user = get_user(user_id)
    if not user.is_active:
        flask.flash(_('Your user account has been deactivated. Please contact an administrator.'), 'error')
        return None
    return user


def _is_url_safe_for_redirect(url: str) -> bool:
    """
    Checks whether a URL is safe for redirecting a user to it.

    URLs are deemed safe if they are absolute URLs on the current domain and
    server path. This is enforced by only allowing URLs that start with the
    absolute server path and only consist of alphanumerical characters or the
    following characters: /=?&_.,+-

    :param url: the URL to check
    :return: whether it is safe to redirect to or not
    """
    server_path = flask.current_app.config.get('SERVER_PATH', '/')
    # ensure the server path starts with a / to avoid relative paths or non-local paths
    if not server_path.startswith('/'):
        server_path = '/' + server_path
    # ensure the server path ends with a / to avoid paths with it as a prefix
    if not server_path.endswith('/'):
        server_path = server_path + '/'
    # prevent double slashes that would change the domain
    if url.startswith('//'):
        return False
    return url.startswith(server_path) and all(c in '/=?&_.,+-%' or c.isalnum() for c in url)


def _redirect_to_next_url(
        next_url: typing.Optional[str] = None
) -> werkzeug.wrappers.response.Response:
    if next_url is None:
        next_url = flask.request.args.get('next', flask.url_for('.index'))
    index_url = flask.url_for('.index')
    if not _is_url_safe_for_redirect(next_url):
        next_url = index_url
    return flask.redirect(next_url)


def _sign_in_impl(is_for_refresh: bool) -> FlaskResponseT:
    if oidc.is_oidc_only_auth_method():
        return flask.redirect(flask.url_for(
            'frontend.oidc_start',
            next=flask.request.args.get('next', '/'),
        ))

    form = SigninForm()
    passkey_form = WebAuthnLoginForm()
    all_credentials = get_all_fido2_passkey_credentials()
    server = get_webauthn_server()

    has_errors = False
    federated_login_token = flask.request.args.get('federated_login_token', None)
    if form.validate_on_submit():
        username = form.username.data
        # enforce lowercase for username to ensure case insensitivity
        # TODO: move this to the model and logic layer
        username = username.lower()
        password = form.password.data
        user = login(username, password)
        if user:
            if not user.is_active:
                flask_login.logout_user()
                flask.flash(_('Your user account has been deactivated. Please contact an administrator.'), 'error')
                return flask.redirect(flask.url_for('frontend.sign_in'))
            if is_for_refresh and user != flask_login.current_user:
                user = None
        if user:
            if not is_for_refresh:
                if user.is_admin and password == 'password':
                    flask.flash(
                        lazy_gettext(
                            'You are using the default admin password from the '
                            'SampleDB documentation. Please change your password '
                            'before making this SampleDB instance available to '
                            'other users.'
                        ),
                        'warning'
                    )
            two_factor_authentication_methods = get_active_two_factor_authentication_methods(user.id)
            if not two_factor_authentication_methods:
                return complete_sign_in(user, is_for_refresh, form.remember_me.data, form.shared_device.data, federated_login_token=federated_login_token)
            flask.session['confirm_data'] = {
                'reason': 'login',
                'user_id': user.id,
                'is_for_refresh': is_for_refresh,
                'remember_me': form.remember_me.data,
                'shared_device': form.shared_device.data,
                'expiration_datetime': (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=5)).strftime('%Y-%m-%d %H:%M:%S'),
                'next_url': flask.request.args.get('next', '/')
            }
            if len(two_factor_authentication_methods) == 1:
                two_factor_authentication_method = two_factor_authentication_methods[0]
                if two_factor_authentication_method.data.get('type') == 'totp':
                    return flask.redirect(flask.url_for('.confirm_totp_two_factor_authentication', method_id=two_factor_authentication_method.id))
                if two_factor_authentication_method.data.get('type') == 'fido2_passkey':
                    return flask.redirect(flask.url_for('.confirm_fido2_passkey_two_factor_authentication', method_id=two_factor_authentication_method.id))
                del flask.session['confirm_data']
                return flask.render_template('two_factor_authentication/unsupported_method.html')
            return flask.render_template(
                'two_factor_authentication/pick.html',
                methods=two_factor_authentication_methods,
            )
        has_errors = True
    elif form.errors:
        has_errors = True
    if passkey_form.validate_on_submit() and flask.current_app.config['ENABLE_FIDO2_PASSKEY_AUTHENTICATION']:
        try:
            credential = server.authenticate_complete(
                flask.session.pop("webauthn_state"),
                credentials=all_credentials,
                response=json.loads(passkey_form.assertion.data),
            )
            user_id = get_user_id_for_fido2_passkey_credential_id(credential.credential_id)
        except Exception:
            user_id = None
        if user_id is not None:
            user = get_user(user_id)
            return complete_sign_in(user, is_for_refresh, passkey_form.remember_me.data, passkey_form.shared_device.data, federated_login_token=federated_login_token)
        flask.flash(_('No user was found for this passkey.'), 'error')
        return flask.redirect(flask.url_for('frontend.sign_in'))

    options, state = server.authenticate_begin(
        credentials=all_credentials,
        user_verification=UserVerificationRequirement.REQUIRED,
    )
    flask.session["webauthn_state"] = state

    federated_logins = get_federated_login_components() if 'federated_login_token' not in flask.request.args else []

    return flask.render_template(
        'sign_in.html',
        form=form,
        is_for_refresh=is_for_refresh,
        has_errors=has_errors,
        passkey_form=passkey_form,
        components=federated_logins,
        options=dict(options),
        saml_user_creation=False,
        federated_login_authentication=bool(flask.session.get('SAMLRequest', None)),
    )


def complete_sign_in(
        user: typing.Optional[User],
        is_for_refresh: bool,
        remember_me: bool,
        shared_device: bool,
        *,
        next_url: typing.Optional[str] = None,
        federated_login_token: typing.Optional[str] = None,
        oidc_id_token_hint: typing.Optional[str] = None,
) -> FlaskResponseT:
    if not user:
        flask.flash(_('Please sign in again.'), 'error')
        flask_login.logout_user()
        return flask.redirect(flask.url_for('.index'))
    if federated_login_token:
        serializer = itsdangerous.URLSafeTimedSerializer(secret_key=flask.current_app.config['SECRET_KEY'], salt='federated-identities')
        try:
            federated_user_id, component_id = serializer.loads(federated_login_token, max_age=150)
        except itsdangerous.BadSignature:
            flask.flash(_('The federated login token expired. No federated identity link was created.'), 'error')

        component = get_component(component_id)
        try:
            identity = get_federated_identity(user_id=user.id, eln_or_fed_user_id=get_user(federated_user_id, component_id).id)
        except (UserDoesNotExistError, FederatedIdentityNotFoundError):
            identity = None

        try:
            if identity is not None and identity.active:
                enable_federated_identity_for_login(identity)
            else:
                create_sampledb_federated_identity(user.id, component=component, fed_id=federated_user_id, update_if_inactive=True, use_for_login=True)
        except FederatedUserInFederatedIdentityError:
            flask.flash(_('The federated user already has a federated identity with a different user.'), 'error')
    if is_for_refresh:
        flask_login.confirm_login()
    else:
        flask_login.login_user(user, remember=remember_me)
        refresh()
    flask.session['using_shared_device'] = shared_device
    if oidc_id_token_hint:
        flask.session['oidc_id_token_hint'] = oidc_id_token_hint
    else:
        flask.session.pop('oidc_id_token_hint', None)
    response = _redirect_to_next_url(next_url)
    response.set_cookie('SAMPLEDB_LAST_ACTIVITY_DATETIME', '', samesite='Lax')
    return response


@frontend.route('/users/me/sign_in', methods=['GET', 'POST'])
def sign_in() -> FlaskResponseT:
    if flask_login.current_user.is_authenticated:
        # if the user was already authenticated, redirect to the index.
        return flask.redirect(flask.url_for('.index'))
    return _sign_in_impl(is_for_refresh=False)


@frontend.route('/users/me/refresh_sign_in', methods=['GET', 'POST'])
@flask_login.login_required
def refresh_sign_in() -> FlaskResponseT:
    if flask_login.current_user.is_authenticated and flask_login.login_fresh():
        # if the login was already fresh, redirect to the next_url or index.
        return _redirect_to_next_url()
    return _sign_in_impl(is_for_refresh=True)


@frontend.route('/users/me/sign_out', methods=['GET', 'POST'])
@flask_login.login_required
def sign_out() -> FlaskResponseT:
    form = SignoutForm()
    if form.validate_on_submit():
        flask_login.logout_user()
        redirect_uri = flask.url_for('.index', _external=True)
        id_token_hint = flask.session.pop('oidc_id_token_hint', None)
        if id_token_hint:
            return flask.redirect(oidc.logout(id_token_hint, redirect_uri))
        return flask.redirect(redirect_uri)
    return flask.render_template('sign_out.html')


@frontend.route('/users/me/shared_device_state')
def shared_device_state() -> FlaskResponseT:
    shared_device_state = False
    # accessing flask_login.current_user triggers the check for a user
    if flask_login.current_user.is_authenticated:
        shared_device_state = bool(flask.g.get('shared_device_state'))
    return flask.jsonify(shared_device_state)


@frontend.route('/users/me/oidc/start')
def oidc_start() -> FlaskResponseT:
    if not oidc.is_oidc_configured():
        flask.flash(_('OIDC is disabled.'), 'error')
        return flask.redirect(flask.url_for('.index'))

    if flask_login.current_user.is_authenticated and flask_login.login_fresh():
        return _redirect_to_next_url()

    url, token = oidc.start_authentication(
        flask.request.args.get('next', flask.url_for('.index')),
    )
    flask.session['oidc_token'] = token
    return flask.redirect(url)


@frontend.route('/users/me/oidc/callback')
def oidc_callback() -> FlaskResponseT:
    if not oidc.is_oidc_configured():
        flask.flash(_('OIDC is disabled.'), 'error')
        return flask.redirect(flask.url_for('.index'))

    if flask_login.current_user.is_authenticated and flask_login.login_fresh():
        return _redirect_to_next_url()

    try:
        result = oidc.handle_authentication(
            flask.request.url,
            flask.session.pop('oidc_token'),
        )
    except Exception:
        flask.flash(_('Login failed. Please contact an administrator.'), 'error')
        flask_login.logout_user()
        return flask.redirect(flask.url_for('.index'))

    user, id_token_hint, next = result

    # While the check in the user_loader would prevent a login, ensure this is
    # handled as soon as possible with a reliable message.
    if not user.is_active:
        flask.flash(_('Your user account has been deactivated. Please contact an administrator.'), 'error')
        return flask.redirect(flask.url_for('.index'))

    return complete_sign_in(
        user,
        is_for_refresh=False,
        remember_me=False,
        shared_device=False,
        next_url=next,
        oidc_id_token_hint=id_token_hint
    )



# coding: utf-8
"""

"""
import typing

import flask
import flask_login
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import StringField, RadioField, PasswordField, SubmitField, IntegerField, HiddenField
from wtforms.validators import DataRequired, Length, Email, ValidationError

from .utils import validate_orcid
from ..logic import authentication


class NewUserForm(FlaskForm):
    name = StringField('Full Name', validators=[DataRequired()])
    email = StringField('Contact Email', validators=[DataRequired("Please enter the contact email."), Email("Please enter your contact email.")])
    password = PasswordField('Password', validators=[DataRequired(), Length(min=3)])
    authentication_method = RadioField('Authentication Method', choices=[('E', 'Email'), ('O', 'Other')], default='O')
    login = StringField('Login', validators=[DataRequired()])
    type = RadioField('Usertype', choices=[('P', 'Person'), ('O', 'Other')], default='O')
    admin = RadioField('Admin', choices=[('0', 'No'), ('1', 'Yes')], default='0')
    submit = SubmitField('Register')


class ChangeUserForm(FlaskForm):
    name = StringField('Full Name', validators=[DataRequired(), Length(min=1)])
    email = StringField('Contact Email', validators=[DataRequired("Please enter the contact email."), Email("Please enter your contact email.")])
    orcid = StringField('ORCID iD')
    affiliation = StringField('Affiliation')
    role = StringField()
    submit = SubmitField('Change Settings')

    def validate_name(self, field: StringField) -> None:
        if flask.current_app.config['ENFORCE_SPLIT_NAMES'] and flask_login.current_user.type.name.lower() == "person":
            name = field.data
            if ', ' not in name[1:-1]:
                raise ValidationError("Please enter your name as: surname, given names.")

    def validate_orcid(self, field: StringField) -> None:
        orcid = field.data
        # accept empty ORCID iDs
        if orcid is None:
            return
        orcid = orcid.strip()
        if not orcid:
            return
        # check ORCID iD syntax
        is_valid, orcid = validate_orcid(orcid)
        if not is_valid:
            raise ValidationError("Please enter a valid ORCID iD.")
        # keep sanitized ORCID iD on success
        field.data = orcid


class LoginForm(FlaskForm):
    username = StringField('Login', validators=[DataRequired()])
    password = PasswordField('Password', validators=[DataRequired(), Length(min=3)])
    submit = SubmitField('Login')


class AuthenticationForm(FlaskForm):
    login = StringField('Login')
    password = PasswordField('Password')
    description = StringField()
    fido2_passkey_credentials = HiddenField()
    authentication_method = RadioField('Authentication Method', choices=[('E', 'Email'), ('L', 'LDAP')], default='E')
    submit = SubmitField('Login')

    def validate(self, *args: typing.Any, **kwargs: typing.Any) -> bool:
        is_valid = bool(super().validate(*args, **kwargs))
        if not is_valid:
            return False
        if self.authentication_method.data in ('E', 'L'):
            if not self.login.data:
                is_valid = False
                self.login.errors.append(_('Please enter at least 1 character.'))
            if not self.password.data or len(self.password.data) < 3:
                is_valid = False
                self.password.errors.append(_('Please enter at least 3 characters.'))
        if self.authentication_method.data == 'FIDO2_PASSKEY':
            if not self.description.data:
                is_valid = False
                self.description.errors.append(_('Please enter at least 1 character.'))
        return is_valid

    def validate_password(self, field: PasswordField) -> None:
        if field.data and len(field.data.encode('utf-8')) > authentication.MAX_BCRYPT_PASSWORD_LENGTH:
            raise ValidationError(_('The password must be at most %(max_bcrypt_password_length)s bytes long.', max_bcrypt_password_length=authentication.MAX_BCRYPT_PASSWORD_LENGTH))


class AuthenticationMethodForm(FlaskForm):
    id = IntegerField('Authentication_method_id', validators=[DataRequired()])
    submit = SubmitField('Remove')



import dataclasses
import typing

from . import errors, components
from .utils import cache
from .. import db, models
from ..models import SciCatExportType


@dataclasses.dataclass(frozen=True)
class ActionType:
    """
    This class provides an immutable wrapper around models.actions.ActionType.
    """
    id: int
    name: typing.Dict[str, str]
    description: typing.Dict[str, str]
    object_name: typing.Dict[str, str]
    object_name_plural: typing.Dict[str, str]
    view_text: typing.Dict[str, str]
    perform_text: typing.Dict[str, str]
    admin_only: bool
    show_on_frontpage: bool
    show_in_navbar: bool
    show_in_object_filters: bool
    enable_labels: bool
    enable_files: bool
    enable_locations: bool
    enable_publications: bool
    enable_comments: bool
    enable_activity_log: bool
    enable_related_objects: bool
    enable_project_link: bool
    enable_instrument_link: bool
    disable_create_objects: bool
    is_template: bool
    order_index: typing.Optional[int]
    usable_in_action_types: typing.List['ActionType'] = dataclasses.field(compare=False)
    fed_id: typing.Optional[int] = None
    component_id: typing.Optional[int] = None
    component: typing.Optional[components.Component] = None
    scicat_export_type: typing.Optional[SciCatExportType] = None

    # make fixed IDs available from wrapper
    SAMPLE_CREATION = models.ActionType.SAMPLE_CREATION
    MEASUREMENT = models.ActionType.MEASUREMENT
    SIMULATION = models.ActionType.SIMULATION
    TEMPLATE = models.ActionType.TEMPLATE

    @classmethod
    def from_database(
            cls,
            action_type: models.ActionType,
            previously_wrapped_action_types: typing.Optional[typing.Dict[int, 'ActionType']] = None
    ) -> 'ActionType':
        if previously_wrapped_action_types is None:
            previously_wrapped_action_types = {}
        wrapped_action_type = ActionType(
            id=action_type.id,
            name=action_type.name,
            description=action_type.description,
            object_name=action_type.object_name,
            object_name_plural=action_type.object_name_plural,
            view_text=action_type.view_text,
            perform_text=action_type.perform_text,
            admin_only=action_type.admin_only,
            show_on_frontpage=action_type.show_on_frontpage,
            show_in_navbar=action_type.show_in_navbar,
            show_in_object_filters=action_type.show_in_object_filters,
            enable_labels=action_type.enable_labels,
            enable_files=action_type.enable_files,
            enable_locations=action_type.enable_locations,
            enable_publications=action_type.enable_publications,
            enable_comments=action_type.enable_comments,
            enable_activity_log=action_type.enable_activity_log,
            enable_related_objects=action_type.enable_related_objects,
            enable_project_link=action_type.enable_project_link,
            enable_instrument_link=action_type.enable_instrument_link,
            disable_create_objects=action_type.disable_create_objects,
            is_template=action_type.is_template,
            order_index=action_type.order_index,
            usable_in_action_types=[],
            fed_id=action_type.fed_id,
            component_id=action_type.component_id,
            component=components.Component.from_database(action_type.component) if action_type.component is not None else None,
            scicat_export_type=action_type.scicat_export_type
        )
        previously_wrapped_action_types[action_type.id] = wrapped_action_type
        # add referenced action types to list after wrapping to support cyclic references
        if action_type.usable_in_action_types:
            for other_action_type in action_type.usable_in_action_types:
                if other_action_type.id in previously_wrapped_action_types:
                    wrapped_action_type.usable_in_action_types.append(previously_wrapped_action_types[other_action_type.id])
                else:
                    wrapped_other_action_type = ActionType.from_database(other_action_type, previously_wrapped_action_types)
                    wrapped_action_type.usable_in_action_types.append(wrapped_other_action_type)
        return wrapped_action_type

    def __repr__(self) -> str:
        return f"<{type(self).__name__}(id={self.id!r})>"


@cache
def check_action_type_exists(
        action_type_id: int
) -> None:
    """
    Check whether an action type with the given action type ID exists.

    :param action_type_id: the ID of an existing action type
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    """
    if not db.session.query(db.exists().where(models.ActionType.id == action_type_id)).scalar():
        raise errors.ActionTypeDoesNotExistError()


def get_action_type(action_type_id: int, component_id: typing.Optional[int] = None) -> ActionType:
    """
    Returns the action type with the given action type ID or composite federation ID.

    :param action_type_id: the ID of an existing action type or ID on other component
        if composite federation ID is used
    :param component_id: optional component ID, if the composite federation ID is used
    :return: the action type
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    """
    if component_id is None:
        action_type = models.ActionType.query.filter_by(id=action_type_id).first()
    else:
        # ensure that the component can be found
        components.check_component_exists(component_id)
        action_type = models.ActionType.query.filter_by(fed_id=action_type_id, component_id=component_id).first()
    if action_type is None:
        raise errors.ActionTypeDoesNotExistError()
    return ActionType.from_database(action_type)


def get_action_types(
        filter_fed_defaults: bool = False
) -> typing.List[ActionType]:
    """
    Return the list of all existing action types.

    :return: the action types
    """
    query = models.ActionType.query

    if filter_fed_defaults:
        query = query.filter(db.or_(models.ActionType.fed_id > 0, models.ActionType.fed_id.is_(None)))

    query = query.order_by(db.nulls_last(models.ActionType.order_index), models.ActionType.id)

    return [
        ActionType.from_database(action_type)
        for action_type in query.all()
    ]


@typing.overload
def create_action_type(
        *,
        admin_only: bool,
        show_on_frontpage: bool,
        show_in_navbar: bool,
        show_in_object_filters: bool,
        enable_labels: bool,
        enable_files: bool,
        enable_locations: bool,
        enable_publications: bool,
        enable_comments: bool,
        enable_activity_log: bool,
        enable_related_objects: bool,
        enable_project_link: bool,
        enable_instrument_link: bool,
        disable_create_objects: bool,
        is_template: bool,
        usable_in_action_type_ids: typing.Sequence[int] = (),
        fed_id: None = None,
        component_id: None = None,
        scicat_export_type: typing.Optional[SciCatExportType] = None
) -> ActionType:
    ...


@typing.overload
def create_action_type(
        *,
        admin_only: bool,
        show_on_frontpage: bool,
        show_in_navbar: bool,
        show_in_object_filters: bool,
        enable_labels: bool,
        enable_files: bool,
        enable_locations: bool,
        enable_publications: bool,
        enable_comments: bool,
        enable_activity_log: bool,
        enable_related_objects: bool,
        enable_project_link: bool,
        enable_instrument_link: bool,
        disable_create_objects: bool,
        is_template: bool,
        usable_in_action_type_ids: typing.Sequence[int] = (),
        fed_id: int,
        component_id: int,
        scicat_export_type: typing.Optional[SciCatExportType] = None
) -> ActionType:
    ...


def create_action_type(
        *,
        admin_only: bool,
        show_on_frontpage: bool,
        show_in_navbar: bool,
        show_in_object_filters: bool,
        enable_labels: bool,
        enable_files: bool,
        enable_locations: bool,
        enable_publications: bool,
        enable_comments: bool,
        enable_activity_log: bool,
        enable_related_objects: bool,
        enable_project_link: bool,
        enable_instrument_link: bool,
        disable_create_objects: bool,
        is_template: bool,
        usable_in_action_type_ids: typing.Sequence[int] = (),
        fed_id: typing.Optional[int] = None,
        component_id: typing.Optional[int] = None,
        scicat_export_type: typing.Optional[SciCatExportType] = None
) -> ActionType:
    """
    Create a new action type.

    :param admin_only: whether actions of this type can only be created by administrators
    :param show_on_frontpage: whether this action type should be shown on the frontpage
    :param show_in_navbar: whether actions of this type should be shown in the navbar
    :param show_in_object_filters: whether this action type should be shown in object filters
    :param enable_labels: whether labels should be enabled for actions of this type
    :param enable_files: whether file uploads should be enabled for actions of this type
    :param enable_locations: whether locations and responsible users should be enabled for actions of this type
    :param enable_publications: whether publications should be enabled for actions of this type
    :param enable_comments: whether comments should be enabled for actions of this type
    :param enable_activity_log: whether the activity log should be enabled for actions of this type
    :param enable_related_objects: whether showing related objects should be enabled for actions of this type
    :param enable_project_link: whether objects created with actions of this type can be linked to a project group
    :param enable_instrument_link: objects created with actions of this type can be linked to an instrument
    :param disable_create_objects: whether the creation of objects with this action type is enabled
    :param is_template: whether this action type can be embedded as a template in actions of other types
    :param usable_in_action_type_ids: sequence of all action types for which objects created with this action type
        should have a "Use in" button
    :param fed_id: action type ID on other database, if action type is imported
    :param component_id: component ID of the component this action type is imported from
    :param scicat_export_type: the SciCat type to use during export, or None
    :return: the created action type
    """
    assert (component_id is None) == (fed_id is None)

    if component_id is not None:
        # ensure that the component can be found
        components.check_component_exists(component_id)

    action_type = models.ActionType(
        admin_only=admin_only,
        show_on_frontpage=show_on_frontpage,
        show_in_navbar=show_in_navbar,
        show_in_object_filters=show_in_object_filters,
        enable_labels=enable_labels,
        enable_files=enable_files,
        enable_locations=enable_locations,
        enable_publications=enable_publications,
        enable_comments=enable_comments,
        enable_activity_log=enable_activity_log,
        enable_related_objects=enable_related_objects,
        enable_project_link=enable_project_link,
        enable_instrument_link=enable_instrument_link,
        disable_create_objects=disable_create_objects,
        is_template=is_template,
        order_index=None,
        usable_in_action_types=[
            models.ActionType.query.filter_by(id=action_type_id).first()
            for action_type_id in usable_in_action_type_ids
        ],
        fed_id=fed_id,
        component_id=component_id,
        scicat_export_type=scicat_export_type
    )
    db.session.add(action_type)
    db.session.commit()
    return ActionType.from_database(action_type)


def update_action_type(
        *,
        action_type_id: int,
        admin_only: bool,
        show_on_frontpage: bool,
        show_in_navbar: bool,
        show_in_object_filters: bool,
        enable_labels: bool,
        enable_files: bool,
        enable_locations: bool,
        enable_publications: bool,
        enable_comments: bool,
        enable_activity_log: bool,
        enable_related_objects: bool,
        enable_project_link: bool,
        enable_instrument_link: bool,
        disable_create_objects: bool,
        is_template: bool,
        usable_in_action_type_ids: typing.Sequence[int] = (),
        scicat_export_type: typing.Optional[SciCatExportType] = None
) -> ActionType:
    """
    Update an existing action type.

    :param action_type_id: the ID of an existing action type
    :param admin_only: whether actions of this type can only be created by administrators
    :param show_on_frontpage: whether this action type should be shown on the frontpage
    :param show_in_navbar: whether actions of this type should be shown in the navbar
    :param show_in_object_filters: whether this action type should be shown in object filters
    :param enable_labels: whether labels should be enabled for actions of this type
    :param enable_files: whether file uploads should be enabled for actions of this type
    :param enable_locations: whether locations and responsible users should be enabled for actions of this type
    :param enable_publications: whether publications should be enabled for actions of this type
    :param enable_comments: whether comments should be enabled for actions of this type
    :param enable_activity_log: whether the activity log should be enabled for actions of this type
    :param enable_related_objects: whether showing related objects should be enabled for actions of this type
    :param enable_project_link: objects created with actions of this type can be linked to a project group
    :param enable_instrument_link: objects created with actions of this type can be linked to an instrument
    :param disable_create_objects: whether the creation of objects with this action type is enabled
    :param is_template: whether this action type can be embedded as a template in actions of other types
    :param usable_in_action_type_ids: sequence of all action types for which objects created with this action type
        should have a "Use in" button
    :param scicat_export_type: the SciCat type to use during export, or None
    :return: the created action type
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    """
    action_type = models.ActionType.query.filter_by(id=action_type_id).first()
    if action_type is None:
        raise errors.ActionTypeDoesNotExistError()
    action_type.admin_only = admin_only
    action_type.show_on_frontpage = show_on_frontpage
    action_type.show_in_navbar = show_in_navbar
    action_type.show_in_object_filters = show_in_object_filters
    action_type.enable_labels = enable_labels
    action_type.enable_files = enable_files
    action_type.enable_locations = enable_locations
    action_type.enable_publications = enable_publications
    action_type.enable_comments = enable_comments
    action_type.enable_activity_log = enable_activity_log
    action_type.enable_related_objects = enable_related_objects
    action_type.enable_project_link = enable_project_link
    action_type.enable_instrument_link = enable_instrument_link
    action_type.disable_create_objects = disable_create_objects
    action_type.is_template = is_template
    usable_in_action_types = [
        models.ActionType.query.filter_by(id=action_type_id).first()
        for action_type_id in usable_in_action_type_ids
    ]
    action_type.usable_in_action_types = [
        other_action_type
        for other_action_type in usable_in_action_types
        if other_action_type is not None
    ]
    action_type.scicat_export_type = scicat_export_type
    db.session.add(action_type)
    db.session.commit()
    return ActionType.from_database(action_type)


def is_usable_in_action_types_table_empty() -> bool:
    """
    Check if the usable in action types table has entries.
    """
    return db.session.query(models.actions.usable_in_action_types_table).first() is None


def set_action_types_order(action_type_id_list: typing.List[int]) -> None:
    """
    Sets the `order_index` for all action types in the index_list, therefore the order in the list is used.

    :param action_type_id_list: list of action type ids
    """
    for i, action_type_id in enumerate(action_type_id_list):
        action_type = models.ActionType.query.filter_by(id=action_type_id).first()
        if action_type is not None:
            action_type.order_index = i
    db.session.commit()


def add_action_type_to_order(action_type: ActionType) -> None:
    """
    Insert an action type into the current sort order.

    :param action_type: the action type to insert
    """
    if action_type.order_index is not None:
        # action type already has a place in the current sort order
        return
    action_types_before: typing.List[ActionType] = []
    action_types_after: typing.List[ActionType] = []
    action_types = [
        ActionType.from_database(other_action_type)
        for other_action_type in models.ActionType.query.filter(models.ActionType.order_index != db.null()).order_by(models.ActionType.order_index).all()
    ]
    if action_types:
        # check if local action types are listed before imported action types
        local_action_types = [
            other_action_type
            for other_action_type in action_types
            if (other_action_type.fed_id is None)
        ]
        imported_action_types = [
            other_action_type
            for other_action_type in action_types
            if (other_action_type.fed_id is not None)
        ]
        local_before_imported = (max(
            typing.cast(int, other_action_type.order_index)
            for other_action_type in local_action_types
        ) < min(
            typing.cast(int, other_action_type.order_index)
            for other_action_type in imported_action_types
        )) if imported_action_types else True
        if local_before_imported:
            if action_type.fed_id is None:
                action_types = local_action_types
                action_types_after = imported_action_types + action_types_after
            else:
                action_types_before = action_types_before + local_action_types
                action_types = imported_action_types
    if action_types:
        # check if action types are listed in order of their english names
        english_names = [
            other_action_type.name.get('en', '').lower()
            for other_action_type in action_types
        ]
        english_lexicographical_order = english_names == sorted(english_names)
        if english_lexicographical_order:
            english_name = action_type.name.get('en', '').lower()
            action_types_before = action_types_before + [
                other_action_type
                for other_action_type in action_types
                if other_action_type.name.get('en', '').lower() < english_name
            ]
            action_types_after = [
                other_action_type
                for other_action_type in action_types
                if other_action_type.name.get('en', '').lower() > english_name
            ] + action_types_after
            action_types = [
                other_action_type
                for other_action_type in action_types
                if other_action_type.name.get('en', '').lower() == english_name
            ]
    # update order indices
    index_list = [
        other_action_type.id
        for other_action_type in (
            action_types_before + action_types + [action_type] + action_types_after
        )
    ]
    set_action_types_order(index_list)
